unit acAlphaHints;
{$I sDefs.inc}
//{$DEFINE LOGGED}
//{$DEFINE DEVEX2011}

interface

uses
  Windows, Messages, SysUtils, Classes, Controls, Forms, ExtCtrls, Graphics,
  {$IFDEF TNTUNICODE} TntForms, TntControls, {$ENDIF}
  {$IFDEF DELPHI7UP} Types, {$ENDIF}
  {$IFDEF DELPHI_XE2} UITypes, {$ENDIF}
  {$IFNDEF FPC} sHtmlParse, {$ENDIF}
  sGraphUtils, sConst, acntUtils, acPNG, acThdTimer;


{$IFNDEF NOTFORHELP}
const
  AddInterval = 0;
  StepDivider = 2;
  Speed       = 5;

  DefAnimationTime = 120;
{$ENDIF}


type
{$IFNDEF NOTFORHELP}
  TacShowHintEvent = procedure(var   HintStr: String; var   CanShow: Boolean; var HintInfo: {$IFDEF D2009}Controls.{$ENDIF}THintInfo; var Frame: TFrame) of object;
  TacMeasureEvent  = procedure(const HintStr: String; const HintInfo: {$IFDEF D2009}Controls.{$ENDIF}THintInfo; var NewWidth, NewHeight: integer) of object;
  TacPaintEvent    = procedure(const HintStr: String; const HintInfo: {$IFDEF D2009}Controls.{$ENDIF}THintInfo; ARect: TRect; ABmp: TBitmap; var DefaultDraw: Boolean) of object;
  TacMousePosition = (mpLeftTop, mpLeftBottom, mpRightTop, mpRightBottom);
  TacBorderDrawMode = (dmRepeat, dmStretch);
  TsAlphaHints        = class;
  TacHintImage        = class;
  TacHintTemplate     = class;
  TacHintTemplates    = class;
  TacCustomHintWindow = class;
{$IFDEF D2009}
  THintInfo = Controls.THintInfo;
{$ENDIF}


  TacBorderDrawModes = class(TPersistent)
  private
    FTop,
    FLeft,
    FRight,
    FCenter,
    FBottom: TacBorderDrawMode;
    FOwner: TacHintImage;
    procedure SetDrawMode(const Index: Integer; const Value: TacBorderDrawMode);
  protected
    procedure AssignTo(Dest: TPersistent); override;
  public
    constructor Create(AOwner: TacHintImage);
    procedure SaveToStream  (AStream: TStream);
    procedure LoadFromStream(AStream: TStream);
  published
    property Top    : TacBorderDrawMode index 0 read FTop    write SetDrawMode default dmStretch;
    property Left   : TacBorderDrawMode index 1 read FLeft   write SetDrawMode default dmStretch;
    property Bottom : TacBorderDrawMode index 2 read FBottom write SetDrawMode default dmStretch;
    property Right  : TacBorderDrawMode index 3 read FRight  write SetDrawMode default dmStretch;
    property Center : TacBorderDrawMode index 4 read FCenter write SetDrawMode default dmStretch;
  end;


  TacBordersSizes = class(TPersistent)
  private
    function GetInteger(const Index: Integer): integer;
  protected
    FTop,
    FLeft,
    FRight,
    FBottom: integer;
    FOwner: TacHintImage;
    procedure AssignTo(Dest: TPersistent); override;
  public
    constructor Create(AOwner: TacHintImage);
    procedure SetInteger(Index: integer; Value: integer);
    procedure SetValues(ALeft, ATop, ARight, ABottom: integer);
    procedure SaveToStream  (AStream: TStream);
    procedure LoadFromStream(AStream: TStream);
  published
    property Top    : integer index 0 read GetInteger write SetInteger default 0;
    property Left   : integer index 1 read GetInteger write SetInteger default 0;
    property Bottom : integer index 2 read GetInteger write SetInteger default 0;
    property Right  : integer index 3 read GetInteger write SetInteger default 0;
  end;


  TacHintImage = class(TPersistent)
  private
    FOffsetHorz,
    FOffsetVert: integer;

    FShadowSizes,
    FClientMargins,
    FBordersWidths: TacBordersSizes;

    FBorderDrawModes: TacBorderDrawModes;
    function GetImgHeight: integer;
    function GetImgWidth:  integer;
    procedure SetImage        (const Value: TPngGraphic);
    procedure SetImgHeight    (const Value: integer);
    procedure SetImgWidth     (const Value: integer);
    procedure SetOffsetHorz   (const Value: integer);
    procedure SetOffsetVert   (const Value: integer);
    procedure SetClientMargins(const Value: TacBordersSizes);
    procedure SetBordersWidths(const Value: TacBordersSizes);
    procedure SetShadowSizes  (const Value: TacBordersSizes);
  protected
    procedure DefineProperties(Filer: TFiler); override;
    procedure ReadData (Reader: TStream);
    procedure WriteData(Writer: TStream);
    procedure AssignTo(Dest: TPersistent); override;
    function BitmapStored: boolean;
  public
    FImage: TPngGraphic;
    FOwner: TacHintTemplate;
    FImageData: TMemoryStream;
    constructor Create(AOwner: TacHintTemplate);
    destructor Destroy; override;
    procedure StreamChanged;
    procedure ImageChanged;
    procedure SavetoStream  (AStream: TStream);
    procedure LoadFromStream(AStream: TStream);
  published
    property OffsetHorz: integer read FOffsetHorz write SetOffsetHorz default 0;
    property OffsetVert: integer read FOffsetVert write SetOffsetVert default 0;
    property ImageHeight: integer read GetImgHeight write SetImgHeight;
    property ImageWidth:  integer read GetImgWidth  write SetImgWidth;
    property Image: TPngGraphic read FImage write SetImage stored BitmapStored;
    property BorderDrawModes: TacBorderDrawModes read FBorderDrawModes write FBorderDrawModes stored True;
    property ClientMargins: TacBordersSizes read FClientMargins write SetClientMargins;
    property BordersWidths: TacBordersSizes read FBordersWidths write SetBordersWidths;
    property ShadowSizes:   TacBordersSizes read FShadowSizes   write SetShadowSizes;
  end;


  TacHintAlign = (haHorzCenter, haVertCenter);

  TacHintTemplate = class(TCollectionItem)
  private
    FImageDefault,
    FImageRightTop,
    FImageLeftBottom,
    FImageRightBottom: TacHintImage;

    FFont: TFont;
    FName: acString;
    FHintAlign: TacHintAlign;
    procedure SetHintImage(const Index: Integer; const Value: TacHintImage);
    procedure SetFont(const Value: TFont);
  protected
    FOwner: TacHintTemplates;
    procedure AssignTo(Dest: TPersistent); override;
  public
    constructor Create(Collection: TCollection); override;
    function SingleImageUsed: boolean;
    destructor Destroy; override;
  published
    property ImageDefault    : TacHintImage index 0 read FImageDefault     write SetHintImage;
    property Img_LeftBottom  : TacHintImage index 1 read FImageLeftBottom  write SetHintImage;
    property Img_RightBottom : TacHintImage index 2 read FImageRightBottom write SetHintImage;
    property Img_RightTop    : TacHintImage index 3 read FImageRightTop    write SetHintImage;
    property Font: TFont read FFont write SetFont;
    property HintAlign: TacHintAlign read FHintAlign write FHintAlign default haHorzCenter;
    property Name: acString read FName write FName;
  end;


  TacHintTemplates = class(TCollection)
  protected
    FOwner: TsAlphaHints;
    function GetOwner: TPersistent; override;
    procedure AssignTo(Dest: TPersistent); override;
    function GetItem (Index: Integer): TacHintTemplate;
    procedure SetItem(Index: Integer; Value: TacHintTemplate);
  public
    destructor Destroy; override;
    constructor Create(AOwner: TsAlphaHints);
    property Items[Index: Integer]: TacHintTemplate read GetItem write SetItem; default;
  end;


  TsShowTimer = class(TacThreadedTimer)
  public
    x,
    y,
    dx,
    dy,
    StepB: real;

    HintWidth,
    HintHeight: integer;

    Step,
    StepCount,
    Transparency: integer;

    SrcRect: TRect;
    FBmpSize: TSize;
    FBmpTopLeft: TPoint;
    CurrentHintForm: TacCustomHintWindow;

    procedure Init(HintForm: TacCustomHintWindow; w, h: integer);
    procedure IntOnTimer(Sender: TObject);
  end;


  TacShowHintData = record
    Width,
    Height: integer;
    Position: TPoint;
    Caption: acString;
    Control: TControl;
    Image: TBitmap;
    ImageAlignment: TacSide;
  end;


  TacHintState = (hsNormal, hsInit);
{$ENDIF}

{$IFDEF DELPHI_XE3}[ComponentPlatformsAttribute(pidWin32 or pidWin64)]{$ENDIF}
  TsAlphaHints = class(TComponent)
{$IFNDEF NOTFORHELP}
  private
    FMaxWidth,
    FPauseHide,
    FOldHintPause: integer;

    FAnimated,
    FHTMLMode,
    FAutoAlignment: boolean;

    FHintPos: TPoint;
    FOnPaint: TacPaintEvent;
    FTextAlignment: TAlignment;
    FOnMeasure: TacMeasureEvent;
    TmpShowData: TacShowHintData;
    FOnShowHint: TacShowHintEvent;
    FDefaultMousePos: TacMousePosition;
{$IFNDEF ACHINTS}
    FActive,
    FUseSkinData: boolean;

    FOnChange,
    FOnHideHint: TNotifyEvent;

    FTemplateIndex: integer;
    FTemplateName: TacStrValue;
    FSkinSection: TsSkinSection;
    FTemplates: TacHintTemplates;
    FHandleDisabledCtrls: boolean;
{$ENDIF}
    function GetAnimated: boolean;
{$IFNDEF ACHINTS}
    procedure SetActive       (const Value: boolean);
    procedure SetSkinData     (const Value: boolean);
    procedure SetPauseHide    (const Value: integer);
{$ENDIF}
    procedure SetTemplates    (const Value: TacHintTemplates);
    procedure SetTemplateName (const Value: TacStrValue);
    procedure SetAutoAlignment(const Value: boolean);
    procedure UpdateHWClass;
  protected
    FOnOldShowHint: TShowHintEvent;
    HintTimeCtrl: TControl;

    HideTimer,
    CheckTimer: TacThreadedTimer;

    FCacheBmp,
    OldAlphaBmp: TBitmap;

    Blend: real;
    CustomFont: TFont;
    State: TacHintState;
    ShowTimer: TsShowTimer;
    CmpChanging, ShadowCalculated: boolean;
    procedure AssignTo(Dest: TPersistent); override;
    procedure ResetHintInfo;
    procedure OnHideTimer (Sender: TObject);
    procedure OnCtrlTimer (Sender: TObject);
    procedure OnCheckTimer(Sender: TObject);
  public
    NewWindow,
    AnimWindow: TForm;
    FHintLocation: TPoint;

    NewBounds: TRect;
    HintFrame: TFrame;
    HintWindowHandle: THandle;
    HintCreating, HintShowing: boolean;
    CurrentHintInfo: {$IFDEF D2009}Controls.{$ENDIF}THintInfo;
    HintWindow: {$IFDEF TNTUNICODE}TTntCustomHintWindow{$ELSE}THintwindow{$ENDIF};

    procedure OnShowHintApp(var HintStr: string; var CanShow: Boolean; var HintInfo: {$IFDEF D2009}Controls.{$ENDIF}THintInfo);
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Loaded; override;
    procedure Changed;
    procedure CheckAutoHintPos;
    procedure InitTemplate;
    procedure AfterConstruction; override;
    function CreateHintBmp(const HintText: string; AFont: TFont; var ResultBmp: TBitmap): TRect;
    procedure ShowHint(TheControl: TControl; const HintText: string; HideTime: Cardinal = 0); overload;
    procedure ShowHint(Position:   TPoint;   const HintText: string; HideTime: Cardinal = 0); overload;
    procedure ShowHint(ShowHintData: TacShowHintData;                HideTime: Cardinal = 0); overload;

//    procedure ShowHintPaused(Position: TPoint; const HintText: string; PauseTime: Cardinal; HideTime: Cardinal = 0);

    procedure HideHint;
    procedure HideAll;
    procedure KillTimer;
    procedure RepaintHint;
    function Skinned: boolean;
    function GetSkinIndex: integer;
    procedure SaveToStream(AStream: TMemoryStream);
    procedure SaveToFile(const FileName: string);
    procedure StartCheckTimer;
    procedure StartHideTimer(HideTime: Cardinal = 0);

    function LoadFromStream(AStream: TMemoryStream): TacHintTemplate;
    function LoadFromFile(const FileName: string): TacHintTemplate;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property DefaultMousePos: TacMousePosition read FDefaultMousePos write FDefaultMousePos default mpLeftTop;
    property IsHintShowing: Boolean read HintShowing;
    property TemplateIndex: integer read FTemplateIndex;
  published
{$ENDIF}
    property Active: boolean read FActive write SetActive default True;
    property AutoAlignment: boolean read FAutoAlignment write SetAutoAlignment default False;
    property Animated: boolean read GetAnimated write FAnimated default True;
    property MaxWidth: integer read FMaxWidth write FMaxWidth default 120;
    property HandleDisabledCtrls: boolean read FHandleDisabledCtrls write FHandleDisabledCtrls default False;
    property Templates: TacHintTemplates read FTemplates write SetTemplates;
    property TemplateName: TacStrValue read FTemplateName write SetTemplateName;
    property TextAlignment: TAlignment read FTextAlignment write FTextAlignment default taCenter;
    property HTMLMode: boolean read FHTMLMode write FHTMLMode default False;
    property PauseHide: integer read FPauseHide write SetPauseHide default 5000;
{$IFNDEF ACHINTS}
    property SkinSection: TsSkinSection read FSkinSection write FSkinSection;
    property UseSkinData: boolean read FUseSkinData write SetSkinData default False;
{$ENDIF}
    property OnShowHint: TacShowHintEvent read FOnShowHint write FOnShowHint;
    property OnHideHint: TNotifyEvent read FOnHideHint write FOnHideHint;
    property OnMeasure: TacMeasureEvent read FOnMeasure write FOnMeasure;
    property OnPaint: TacPaintEvent read FOnPaint write FOnPaint;
  end;


{$IFNDEF NOTFORHELP}
  TacCustomHintWindow = class({$IFDEF TNTUNICODE}TTntCustomHintWindow{$ELSE}THintWindow{$ENDIF})
  private
    Text: string;
    procedure PrepareMask;
    procedure WMEraseBkGND(var Message: TWMPaint);   message WM_ERASEBKGND;
    procedure WMNCPaint   (var Message: TWMPaint);   message WM_NCPaint;
    procedure WMDestroy   (var Message: TWMDestroy); message WM_DESTROY;
  protected
    BodyBmp,
    MaskBmp,
    AlphaBmp: TBitmap;

    rgn: hrgn;
    SkinIndex: integer;
    AlphaUsed: boolean;
    OldSDProc: TWndMethod;
    OldFrameProc: TWndMethod;
    FMousePos: TacMousePosition;
    procedure CreateAlphaBmp(const Width, Height: integer); virtual;
    procedure CreateParams(var Params: TCreateParams); override;
{$IFDEF DEVEX2011}
    procedure CreateWnd; override;
{$ENDIF}
    function GetMask: TBitmap;
    function GetBody: TBitmap;
    function GetMousePosition: TPoint;
    function MainRect: TRect;
    function SkinMargin(Border: byte): integer;
    procedure TextOut(Bmp: TBitmap);
    function FullLayer: boolean;
    function ShadowSizes: TRect;
    procedure WndProc(var Message: TMessage); override;
    procedure FrameWndProc(var Message: TMessage);
    procedure Paint; override;

    function CreateHintWindow(R: TRect): TForm;
    procedure DoDestroy; virtual;
    procedure OnCloseHint;
    function GetArrowPos(var Rect: TRect; const mPos: TPoint): TacMousePosition;
    procedure UpdateWnd(w, h: integer);
  public
    procedure ActivateHint(Rect: TRect; const AHint: string); {IFNDEF TNTUNICODE} override; {ENDIF}
    function CalcHintRect(MaxWidth: Integer; const AHint: string; AData: Pointer): TRect; {IFNDEF TNTUNICODE} override; {ENDIF}
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  end;


var
  Manager: TsAlphaHints;
{$ENDIF}

{$IFDEF ACHINTS}
procedure Register;
{$ENDIF}

implementation

uses
  math,
  {$IFNDEF WIN64} acZLibEx, {$ELSE} ZLib, {$ENDIF}
  {$IFNDEF ACHINTS} sVclUtils, sMessages, sSkinProps, sSkinManager, {$ENDIF}
  {$IFDEF LOGGED} sDebugMsgs, {$ENDIF}
  {$IFDEF DEVEX2011} cxGrid, cxGridCustomView, cxGridTableView, {$ENDIF}
  sAlphaGraph, sCommonData, sStyleSimply;

{$R acHints.res}

var
  Image:           TacHintImage    = nil;
  acHintWindow:    THintWindow     = nil;
  DefaultTemplate: TacHintTemplate = nil;
  Template:        TacHintTemplate = nil;
  acLocalUpd:      boolean         = False;


type
  TacPngHintWindow = class(TacCustomHintWindow)
  public
    constructor Create(AOwner: TComponent); override;
    procedure UpdateImageTemplate;
    procedure DoDestroy; override;
    procedure CreateAlphaBmp(const Width, Height: integer); override;
    function CalcHintRect(MaxWidth: Integer; const AHint: string; AData: Pointer): TRect; {IFNDEF TNTUNICODE} override; {ENDIF}
  end;

  TacTrackHintWindow = class(TacPngHintWindow);


{$IFDEF ACHINTS}
procedure Register;
begin
  RegisterComponents('AlphaTools', [TsAlphaHints]);
end;
{$ENDIF}


function SkinBorderWidth: integer;
begin
  Result := 4;
  if DefaultManager <> nil then
    Result := DefaultManager.ScaleInt(Result);
end;


procedure UpdateFontSize(AFont: TFont; CustomFont: boolean);
var
  OrigSize: integer;
begin
  if CustomFont then
    OrigSize := AFont.Height
  else
    OrigSize := -12;

  if (DefaultManager <> nil) and (DefaultManager.Options.ScaleMode <> smVCL) then
    AFont.Height := DefaultManager.ScaleInt(OrigSize)
  else
    if (Manager <> nil) and (Manager.CurrentHintInfo.HintControl <> nil) then
      AFont.Height := MulDiv(OrigSize, GetControlPPI(Manager.CurrentHintInfo.HintControl), 96)
    else
      if Application.MainForm <> nil then
        AFont.Height := MulDiv(OrigSize, GetControlPPI(Application.MainForm), 96)
end;


constructor TsAlphaHints.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FOnShowHint := nil;
  FTemplateIndex := -1;
  FHandleDisabledCtrls := False;
  FCacheBmp := CreateBmp32;
  HintFrame := nil;
  FHTMLMode := False;
  FHintLocation.X := 0;
  FHintLocation.Y := 0;
  FPauseHide := 5000;
  FMaxWidth := 120;
  FDefaultMousePos := mpLeftTop;
  FTextAlignment := taCenter;
  FTemplates := TacHintTemplates.Create(Self);
  CustomFont := nil;
  HintShowing := False;
  HintCreating := False;
  FAnimated := True;
  FAutoAlignment := False;
  FActive := True;
  AnimWindow := nil;
  State := hsNormal;
  OldAlphaBmp := nil;
  Blend := 0;
end;


function TsAlphaHints.CreateHintBmp(const HintText: string; AFont: TFont; var ResultBmp: TBitmap): TRect;
var
  SavedRect, HintRect: TRect;
  SavedPos, SavedLocation: TPoint;
  iMaxWidth: integer;
  hw: TacTrackHintWindow;
begin
  SavedRect := NewBounds;
  SavedLocation := FHintLocation;
  SavedPos := FHintPos;
  hw := TacTrackHintWindow.Create(Self);
  with hw do begin
    hw.Text := HintText;
    Manager.CustomFont := AFont;
    iMaxWidth := FMaxWidth * sSkinManager.CurrentPPI(nil, hw) div 96;
    HintRect := CalcHintRect(iff(iMaxWidth < 1, Screen.Width, iMaxWidth), HintText, nil);

    FCacheBmp.PixelFormat := pf32bit;
    FCacheBmp.Width := WidthOf(HintRect);
    FCacheBmp.Height := HeightOf(HintRect);

    hw.Width := FCacheBmp.Width - (ShadowSizes.Left + ShadowSizes.Right);
    hw.Height := FCacheBmp.Height - (ShadowSizes.Top + ShadowSizes.Bottom);

    Manager.NewBounds := HintRect;
    hw.CreateAlphaBmp(FCacheBmp.Width, FCacheBmp.Height);
    CustomFont := nil;
    ResultBmp := AlphaBmp;
    AlphaBmp := nil;
    Result := ShadowSizes;
  end;
  FHintPos := SavedPos;
  NewBounds := SavedRect;
  FHintLocation := SavedLocation;
end;


destructor TsAlphaHints.Destroy;
begin
  if Assigned(FOnOldShowHint) then
    Application.OnShowHint := FOnOldShowHint
  else
    Application.OnShowHint := nil;

  HintWindowClass := THintWindow;
  UpdateHWClass;
  FreeAndNil(FTemplates);
  FreeAndNil(FCacheBmp);

  HideTimer.Free;
  if ShowTimer <> nil then
    FreeAndNil(ShowTimer);

  if AnimWindow <> nil then
    FreeAndNil(AnimWindow);

  if NewWindow <> nil then
    FreeAndNil(NewWindow);

  if CheckTimer <> nil then
    CheckTimer.Free;

  if OldAlphaBmp <> nil then
    FreeAndNil(OldAlphaBmp);
    
  inherited;
end;


procedure TsAlphaHints.Loaded;
begin
  inherited;
  if not (csDesigning in ComponentState) and (Tag <> ExceptTag) then begin
    Application.HintHidePause := FPauseHide;
    FHintPos := Point(-1, -1);
    if Assigned(Application.OnShowHint) then
      FOnOldShowHint := Application.OnShowHint;

    Application.OnShowHint := OnShowHintApp;
  end;
{$IFNDEF ACHINTS}
  if FSkinSection = '' then
    FSkinSection := s_Hint;
{$ENDIF}
  UpdateHWClass;
end;


procedure TsAlphaHints.OnShowHintApp(var HintStr: string; var CanShow: Boolean; var HintInfo: {$IFDEF D2009}Controls.{$ENDIF}THintInfo);
{$IFDEF DEVEX2011}
var
  AHitTest: TcxCustomGridHitTest;
  APoint: TPoint;
  Grid: TcxGrid;
{$ENDIF}
begin
  if Assigned(FOnOldShowHint) then
    FOnOldShowHint(HintStr, CanShow, HintInfo);

{$IFDEF DEVEX2011}
  if HintInfo.HintControl is TcxGrid then begin
    Grid := TcxGrid(HintInfo.HintControl);
    APoint := Grid.ActiveView.Site.ScreenToClient(acMousePos);
    AHitTest := Grid.ActiveView.Site.ViewInfo.GetHitTest(APoint);
    if AHitTest.HitTestCode = 202 {htColumnHeader} then
      HintInfo.HintPos.Y := HintInfo.HintPos.Y + TcxGridColumnHeaderHitTest(AHitTest).ViewInfo.Height;
  end;
{$ENDIF}

  FHintPos := HintInfo.HintPos;
  if HandleDisabledCtrls and (HintTimeCtrl <> nil) and not HintTimeCtrl.Enabled then
    HintInfo.HintControl := HintTimeCtrl;

  if FAutoAlignment then begin
    CurrentHintInfo := HintInfo;
    InitTemplate;
    CheckAutoHintPos;
    HintInfo.HintPos := FHintPos;
  end;

  if Assigned(FOnShowHint) then begin
    ResetHintInfo;
    FOnShowHint(HintStr, CanShow, HintInfo, HintFrame);
    if (HintFrame <> nil) and (DefaultManager <> nil) then
      DefaultManager.UpdateScale(HintFrame);
  end
  else
    inherited;

  CurrentHintInfo := HintInfo;
  if (FHintPos.x <> HintInfo.HintPos.x) or (FHintPos.y <> HintInfo.HintPos.y) then
    FHintPos := HintInfo.HintPos;
end;


procedure TsAlphaHints.ShowHint(TheControl: TControl; const HintText: string; HideTime: Cardinal = 0);
var
  b: boolean;
  HL: TPoint; // Hint location
  HintRect: TRect;
  NewText: string;
  iMaxWidth: integer;
  HintInfo: {$IFDEF D2009}Controls.{$ENDIF}THintInfo;
begin
  if not HintCreating then begin
    try
      HintCreating := True;
      if (ShowTimer <> nil) and ShowTimer.Enabled then begin
        FreeAndNil(AnimWindow);
        AnimWindow := NewWindow;
        NewWindow := nil;
        FreeAndNil(ShowTimer);
      end;
      if Assigned(TheControl) then begin
        HintTimeCtrl := TheControl;
        HL := MkPoint;
        // Does hint need to be created?
        if not Assigned(HintWindow) then begin
          HintWindow := TacPngHintWindow.Create(Self);
          HintWindowHandle := HintWindow.Handle;
        end;
        // Was hint creation successful?
        if Assigned(HintWindow) then
          with TacCustomHintWindow(HintWindow) do begin
            iMaxWidth := FMaxWidth * sSkinManager.CurrentPPI(nil, HintWindow) div 96;
            HintRect := CalcHintRect(iff(iMaxWidth < 1, Screen.Width, iMaxWidth), HintText, nil);

            FHintLocation := TheControl.ClientToScreen(HL);
            FHintLocation := Point(FHintLocation.x + TheControl.Width div 2, FHintLocation.Y + TheControl.Height div 3);
            if Assigned(OnShowHint) then begin
              b := True;
              HintFrame := nil;
              HintInfo.HintControl := TheControl;
              HintInfo.HintPos := FHintLocation;
              // Save orig. props
              Text := HintText;
              ResetHintInfo;
              NewText := HintText;
              OnShowHint(NewText, b, HintInfo, HintFrame);
              if (HintFrame <> nil) and (DefaultManager <> nil) then
                DefaultManager.UpdateScale(HintFrame);

              if (NewText <> HintText) or (HintFrame <> nil) then
                HintRect := CalcHintRect(iff(iMaxWidth < 1, Screen.Width, iMaxWidth), NewText, NIL);

              CurrentHintInfo := HintInfo;
              FHintLocation := HintInfo.HintPos;
            end
            else
              if HintText <> '' then begin
                NewText := HintText;
                HintRect := CalcHintRect(iff(iMaxWidth < 1, Screen.Width, iMaxWidth), NewText, NIL);
              end;

            ActivateHint(HintRect, NewText);
            Manager.FHintLocation := Point(0, 0);
          end;

        StartHideTimer(HideTime);
      end;
    finally
      HintCreating := False;
    end;
  end;
end;


procedure TsAlphaHints.ShowHint(ShowHintData: TacShowHintData; HideTime: Cardinal = 0);
var
  HL: TPoint; // Hint location
  HintRect: TRect;
  NewText: string;
begin
  if not HintCreating then begin
    HintCreating := True;
    try
      if (ShowTimer <> nil) and ShowTimer.Enabled then begin
        FreeAndNil(AnimWindow);
        AnimWindow := NewWindow;
        NewWindow := nil;
        FreeAndNil(ShowTimer);
      end;
      if Manager <> nil then begin
        if ShowHintData.Control <> nil then
          HintTimeCtrl := ShowHintData.Control
        else
          HintTimeCtrl := CtrlUnderMouse;

        TmpShowData := ShowHintData;
        if TmpShowData.Control <> nil then
          HL := MkPoint
        else
          HL := ShowHintData.Position;
        // Does hint need to be created?
        if not Assigned(HintWindow) then begin
  {$IFDEF DEVEX2011}
          HintWindow := TacPngHintWindow.CreateParented(Application.Handle);
          acHintWindow := HintWindow;
  {$ELSE}
          HintWindow := TacPngHintWindow.Create(Self);
  {$ENDIF}
          HintWindowHandle := HintWindow.Handle;
        end;
        // Was hint creation successful?
        if Assigned(HintWindow) then
          with TacCustomHintWindow(HintWindow) do begin
            FMousePos := Manager.FDefaultMousePos;
            HintRect := Rect(HL.X, HL.Y, HL.X + TmpShowData.Width, HL.Y + TmpShowData.Height);
            FHintLocation := HL;
            HintRect := CalcHintRect(TmpShowData.Width, TmpShowData.Caption, @TmpShowData);
            ActivateHint(HintRect, NewText);
            Manager.FHintLocation := Point(0, 0);
          end;

        StartHideTimer(HideTime);
        FillMemory(@TmpShowData, SizeOf(TmpShowData), 0);
      end;
    finally
      HintCreating := False;
    end;
  end;
end;


procedure TsAlphaHints.HideHint;
begin
  HintFrame := nil;
  if IsWindow(HintWindowHandle) then
    FreeAndNil(HintWindow);

  HintWindow := nil;
  HintWindowHandle := 0;
  if not HandleDisabledCtrls or (HintTimeCtrl <> niL) and HintTimeCtrl.Enabled then
    HintTimeCtrl := nil; // Desc 12
end;


procedure TsAlphaHints.HideAll;
begin
  ResetHintInfo;
  HideHint;
  FreeAndNil(AnimWindow);
  FreeAndNil(NewWindow);
end;


procedure TsAlphaHints.AfterConstruction;
begin
  inherited;
{$IFNDEF ACHINTS}
  if FSkinSection = '' then
    FSkinSection := s_Hint;
{$ENDIF}
end;


function TsAlphaHints.GetAnimated: boolean;
begin
  Result := FAnimated;
end;


function TsAlphaHints.GetSkinIndex: integer;
begin
  ShadowCalculated := False;
  if UseSkinData and Skinned then begin
    Result := DefaultManager.GetSkinIndex(Manager.SkinSection);
    if Result >= 0 then
      ShadowCalculated := DefaultManager.CommonSkinData.gd[Result].OuterMode > 2;
  end
  else
    Result := -1;
end;


function TsAlphaHints.Skinned: boolean;
begin
{$IFNDEF ACHINTS}
  if Assigned(Manager) and Assigned(DefaultManager) and UseSkinData then
    Result := (Manager.SkinSection <> '') and (DefaultManager.GetSkinIndex(Manager.SkinSection) >= 0)
  else
{$ENDIF}
    Result := False
end;


{$IFNDEF ACHINTS}
procedure TsAlphaHints.SetSkinData(const Value: boolean);
begin
  FUseSkinData := Value;
end;
{$ENDIF}


procedure TsAlphaHints.SetPauseHide(const Value: integer);
begin
  if FPauseHide <> Value then begin
    FPauseHide := Value;
    if not (csDesigning in ComponentState) and (Tag <> ExceptTag) then
      Application.HintHidePause := FPauseHide;
  end;
end;


procedure TsAlphaHints.SetActive(const Value: boolean);
begin
  FActive := Value;
  if not FActive and not (csDestroying in ComponentState) then
    HideAll;

  UpdateHWClass;
end;


procedure TsAlphaHints.UpdateHWClass;
begin
  if not (csDesigning in ComponentState) and (Tag <> ExceptTag) {Edit mode for component in editor} then
    if FActive and acLayered and not (csDestroying in ComponentState) then begin
      Manager := Self;
      HintWindowClass := TacPngHintWindow
    end
    else
      if Manager = Self then begin
        Manager := nil;
        HintWindowClass := THintWindow;
      end
      else
        if Manager <> nil then
          Manager.FHintLocation := MkPoint;
end;


procedure TsAlphaHints.SetTemplates(const Value: TacHintTemplates);
begin
  FTemplates.Assign(Value);
end;


procedure TsAlphaHints.SetTemplateName(const Value: TacStrValue);
var
  i: integer;
begin
  FTemplateIndex := -1;
  FTemplateName := Value;
  for i := 0 to Templates.Count - 1 do
    if UpperCase(Templates[i].Name) = UpperCase(Value) then begin
      FTemplateIndex := i;
      Break;
    end;
end;


procedure TsAlphaHints.SetAutoAlignment(const Value: boolean);
begin
  FAutoAlignment := Value;
  FHintLocation := Point(0, 0);
end;


procedure TsAlphaHints.AssignTo(Dest: TPersistent);
var
  DstCmp: TsAlphaHints;
begin
  if Dest <> nil then begin
    DstCmp := TsAlphaHints(Dest);
    DstCmp.Templates.Clear;
    Templates.AssignTo(DstCmp.Templates);
    DstCmp.TemplateName := TemplateName;
    DstCmp.Animated := Animated;
    DstCmp.AutoAlignment := AutoAlignment;
    DstCmp.DefaultMousePos := DefaultMousePos;
    DstCmp.HTMLMode := HTMLMode;
    DstCmp.PauseHide := PauseHide;
    DstCmp.SkinSection := SkinSection;
    DstCmp.UseSkinData := UseSkinData;
  end
  else
    inherited;
end;


procedure TsAlphaHints.Changed;
begin
  if not (csLoading in ComponentState) and Assigned(FOnChange) then
    FOnChange(Self);
end;


procedure TsAlphaHints.CheckAutoHintPos;
var
  R: TRect;

  function GetControlRect(Ctrl: TControl): TRect;
  begin
    if Ctrl is TWinControl then
      GetWindowRect(TWinControl(Ctrl).Handle, Result)
    else begin
      GetWindowRect(Ctrl.Parent.Handle, Result);
      inc(Result.Top,  Ctrl.Top);
      inc(Result.Left, Ctrl.Left);
      Result.Right  := Result.Left + Ctrl.Width;
      Result.Bottom := Result.Top  + Ctrl.Height;
    end;
  end;

begin
  if (Manager <> nil) and Manager.AutoAlignment then begin
    if (Manager.CurrentHintInfo.HintControl = nil) then
      if Manager.HandleDisabledCtrls and (Manager.HintTimeCtrl <> nil) then
        Manager.CurrentHintInfo.HintControl := Manager.HintTimeCtrl
      else
        Exit
    else
      if Manager.HandleDisabledCtrls and (Manager.HintTimeCtrl <> nil) then
        Manager.CurrentHintInfo.HintControl := Manager.HintTimeCtrl;

    R := GetControlRect(Manager.CurrentHintInfo.HintControl);
    if Template <> nil then begin
      if Template.HintAlign = haHorzCenter then begin
        Manager.FHintPos.X := R.Left + WidthOf(R) div 2;
        if (Image = Template.FImageRightBottom) or (Image = Template.FImageLeftBottom) then
          Manager.FHintPos.Y := R.Top
        else
          Manager.FHintPos.Y := R.Bottom;
      end
      else begin
        Manager.FHintPos.Y := R.Top + HeightOf(R) div 2;
        if (Image = Template.FImageRightTop) or (Image = Template.FImageRightBottom) then
          Manager.FHintPos.X := R.Left
        else
          Manager.FHintPos.X := R.Right;
      end;
      if (Image = Template.FImageRightTop) or (Image = Template.FImageRightBottom) then
        dec(Manager.FHintPos.X, Image.ShadowSizes.Right)
      else
        inc(Manager.FHintPos.X, Image.ShadowSizes.Left);

      if (Image = Template.FImageRightBottom) or (Image = Template.FImageLeftBottom) then
        dec(Manager.FHintPos.Y, Image.ShadowSizes.Bottom)
      else
        inc(Manager.FHintPos.Y, Image.ShadowSizes.Top);

      inc(Manager.FHintPos.X, Image.OffsetHorz);
      inc(Manager.FHintPos.Y, Image.OffsetVert);
    end
    else begin
      Manager.FHintPos.X := R.Left + WidthOf(R) div 2;
      Manager.FHintPos.Y := R.Bottom;
    end;
    FHintLocation := Manager.FHintPos;
  end;
end;


procedure TsAlphaHints.InitTemplate;
begin
  if (TemplateIndex >= 0) and (TemplateIndex < Templates.Count) then
    Template := Templates[FTemplateIndex]
  else
    Template := DefaultTemplate;

  if (Self.DefaultMousePos <> mpLeftTop) and not Template.SingleImageUsed then
    case Self.DefaultMousePos of
      mpLeftBottom:  Image := Template.Img_LeftBottom;
      mpRightTop:    Image := Template.Img_RightTop;
      mpRightBottom: Image := Template.FImageRightBottom;
    end
  else
    Image := Template.FImageDefault;
end;


procedure TsAlphaHints.ShowHint(Position: TPoint; const HintText: string; HideTime: Cardinal = 0);
var
  b: boolean;
  HintRect: TRect;
  NewText: string;
  TmpCtrl: TControl;
  iMaxWidth: integer;
  HintInfo: {$IFDEF D2009}Controls.{$ENDIF}THintInfo;
begin
  if not HintCreating then begin
    try
      HintCreating := True;
      if (ShowTimer <> nil) and ShowTimer.Enabled then begin
        FreeAndNil(AnimWindow);
        AnimWindow := NewWindow;
        NewWindow := nil;
        FreeAndNil(ShowTimer);
      end;
      TmpCtrl := CtrlUnderMouse;
      if HintTimeCtrl <> TmpCtrl then begin
        HideAll;
        HintTimeCtrl := TmpCtrl;
      end;
      if (HintWindowHandle = 0) or not IsWindow(HintWindowHandle) then
        HintWindow := nil;

      // Does hint need to be created?
      if not Assigned(HintWindow) then begin
        HintWindow := TacPngHintWindow.Create(Self);
        HintWindowHandle := HintWindow.Handle;
      end;
      // Was hint creation successful?
      if Assigned(HintWindow) then begin
        with TacCustomHintWindow(HintWindow) do begin
          FHintLocation := Position;
          iMaxWidth := FMaxWidth * sSkinManager.CurrentPPI(nil, HintWindow) div 96;
          HintRect := CalcHintRect(iff(iMaxWidth < 1, Screen.Width, iMaxWidth), HintText, nil);
          if Assigned(OnShowHint) then begin
            b := True;
            HintFrame := nil;
            HintInfo.HintControl := nil;
            HintInfo.HintPos := FHintLocation;
            // Save orig. props
            Text := HintText;
            ResetHintInfo;
            NewText := HintText;
            OnShowHint(NewText, b, HintInfo, HintFrame);
            if (HintFrame <> nil) and (DefaultManager <> nil) then
              DefaultManager.UpdateScale(HintFrame);

            if (NewText <> HintText) or (HintFrame <> nil) then
              HintRect := CalcHintRect(iff(iMaxWidth < 1, Screen.Width, iMaxWidth), NewText, nil);

            FHintLocation := HintInfo.HintPos;
          end
          else
            if HintText <> '' then begin
              NewText := HintText;
              HintRect := CalcHintRect(iff(iMaxWidth < 1, Screen.Width, iMaxWidth), NewText, nil);
            end;

          ActivateHint(HintRect, NewText);
          FHintLocation := Point(0, 0);
        end;
        StartHideTimer(HideTime);
      end;
    finally
      HintCreating := False;
    end;
  end;
end;


procedure TsAlphaHints.KillTimer;
begin
  if (ShowTimer <> nil) and ShowTimer.Enabled then // If animation is not finished
//    ShowTimer.Enabled := False; // May be removed?
    FreeAndNil(ShowTimer);
end;


const
  pArray: array [boolean, boolean] of TacMousePosition = ((mpRightBottom, mpLeftBottom), (mpRightTop, mpLeftTop));

procedure TsAlphaHints.RepaintHint;
var
  Rect: TRect;
  mPos: TPoint;
  F: TFrame;
  b: boolean;
  t, l, Auto: boolean;
  h, w: integer;
{$IFDEF DELPHI6UP}
  Monitor: TMonitor;
{$ENDIF}
  iMaxWidth: integer;
  HintInfo: {$IFDEF D2009}Controls.{$ENDIF}THintInfo;

  procedure SetLeft(Value: integer; al: boolean);
  begin
    Rect.Left := Value;
    l := al;
    Auto := True;
  end;

  procedure SetTop(Value: integer; at: boolean);
  begin
    Rect.Top := Value;
    t := at;
    Auto := True;
  end;

begin
  if (acHintWindow <> nil) and HintShowing then begin
    if Assigned(OnShowHint) then begin
      b := True;
      F := HintFrame;
      HintInfo := CurrentHintInfo;
      OnShowHint(HintInfo.HintStr, b, HintInfo, F);
      CurrentHintInfo := HintInfo;
    end;
    if NewWindow <> nil then
      FreeAndNil(NewWindow);

    iMaxWidth := FMaxWidth * sSkinManager.CurrentPPI(nil, HintWindow) div 96;
    Rect := TacCustomHintWindow(acHintWindow).CalcHintRect(iff(iMaxWidth < 1, Screen.Width, iMaxWidth), HintInfo.HintStr, nil);
    Manager.NewBounds := Rect;
    w := WidthOf(Rect);
    h := HeightOf(Rect);
    TacCustomHintWindow(acHintWindow).FMousePos := DefaultMousePos;
    mPos := TacCustomHintWindow(acHintWindow).GetMousePosition;
    OffsetRect(Rect, mPos.X, mPos.Y);
{$IFDEF DELPHI6UP}
    Monitor := Screen.MonitorFromPoint(mPos);
    if Monitor <> nil then begin
      if Rect.Bottom > Monitor.Top + Monitor.Height then
        if Manager.Skinned then
          SetTop(Monitor.Top + Monitor.Height - h, False)
        else
          if Template.SingleImageUsed then
            SetTop(mPos.Y - (Rect.Bottom - (Monitor.Top + Monitor.Height)), False)
          else
            SetTop(mPos.Y - h, False);

      if mPos.Y < Monitor.Top then
        SetTop(Monitor.Top, True);

      if Rect.Right > Monitor.Left + Monitor.Width then
        if Manager.Skinned then
          SetLeft(Monitor.Left + Monitor.Width - w, False)
        else
          if Template.SingleImageUsed then
            SetLeft(mPos.X - (Rect.Right - (Monitor.Left + Monitor.Width)), False)
          else
            SetLeft(mPos.X - w, False);

      if mPos.X < Monitor.Left then
        SetLeft(Monitor.Left, True);
    end;
  {$ELSE}
    if Rect.Bottom > Screen.Height then
      SetTop(mPos.y - h, False);

    if Rect.Top < 0 then
      SetTop(mPos.y, True);

    if Rect.Right > Screen.Width then
      SetLeft(mPos.x - w, False);

    if Rect.Left < 0 then
      SetLeft(mPos.x, True);
  {$ENDIF}
    if Auto then
      TacCustomHintWindow(acHintWindow).FMousePos := pArray[t, l];

//    TacCustomHintWindow(acHintWindow).SetBounds(Rect.Left, Rect.Top, w, h);
    Manager.NewBounds := Classes.Rect(Rect.Left, Rect.Top, Rect.Left + w, Rect.Top + h);
    TacCustomHintWindow(acHintWindow).UpdateWnd(w, h);
  end;
end;


procedure TsAlphaHints.ResetHintInfo;
begin
  if Manager <> nil then
    if [csDesigning, csLoading] * Manager.ComponentState = [] then begin
      Manager.TmpShowData.Control := nil;
      Manager.CurrentHintInfo.HintControl := nil;
      if Assigned(HintFrame) then
        FreeAndNil(HintFrame);

      if Manager <> nil then
        Manager.HintShowing := False;
    end;
end;


procedure TacCustomHintWindow.ActivateHint(Rect: TRect; const AHint: string);
{$IFNDEF ACHINTS}
var
  w, h: integer;
  p: TPoint;
{$ENDIF}
begin
{$IFNDEF ACHINTS}
  if Assigned(Manager) then begin
    Manager.State := hsInit;
    FreeAndNil(Manager.CheckTimer);
    Manager.HintShowing := True;

    if Manager.AnimWindow <> nil then // If moved
      if (Manager.AnimWindow.Width = 0) or (Manager.AnimWindow.Left = 0) and (Manager.AnimWindow.Top = 0) then
        FreeAndNil(Manager.AnimWindow) // Fix for DevEx hints
      else
        if (Manager.ShowTimer <> nil) and Manager.ShowTimer.Enabled then begin // If prev animation is not finished
          Manager.KillTimer;
          if Manager.NewWindow = nil then // If was showing animation
            if AlphaBmp <> nil then begin
              if Manager.OldAlphaBmp <> nil then
                Manager.OldAlphaBmp.Free;

              Manager.OldAlphaBmp := AlphaBmp;
              AlphaBmp := nil;
            end;
        end;

    Manager.KillTimer;
{$IFDEF DEVEX2011}
    Application.ProcessMessages; // Patch for refresh of columns in cxGrid
{$ENDIF}
    if Manager.Active then
      Text := AHint
    else
      Caption := AHint;

    if (Manager.FHintLocation.X = 0) or (Manager.FHintLocation.Y = 0) then
      p := Rect.TopLeft
    else
      p := Manager.FHintLocation;

    w := WidthOf(Rect);
    h := HeightOf(Rect);
    OffsetRect(Rect, p.x - Rect.Left - ShadowSizes.Left, p.y - Rect.Top - ShadowSizes.Top);

    // Show std window as fully transparent
    if HandleAllocated then begin
      SetWindowLong(Handle, GWL_EXSTYLE, GetWindowLong(Handle, GWL_EXSTYLE) or WS_EX_LAYERED or WS_EX_NOACTIVATE);
      SetBounds(Rect.Left, Rect.Top, w, h);
      SetClassLong(Handle, GCL_STYLE, GetClassLong(Handle, GCL_STYLE) and not NCS_DROPSHADOW);
      ShowWindow(Handle, SW_SHOWNOACTIVATE); // Hints are not closed in cxGrids, if window is not shown
    end; // Don't exit if not HandleAllocated, hint is not shown if exit

    FMousePos := GetArrowPos(Rect, p);
    if not Manager.Skinned and (FMousePos in [mpLeftBottom, mpRightBottom]) and not Manager.AutoAlignment then
      dec(Rect.Top, 16);

    Rect.Right := Rect.Left + w;
    Rect.Bottom := Rect.Top + h;
    Manager.FCacheBmp.PixelFormat := pf32bit;
    Manager.FCacheBmp.Width := w;
    Manager.FCacheBmp.Height := h;

    Manager.NewBounds := Rect;
    if Manager.AnimWindow = nil then
      Manager.AnimWindow := CreateHintWindow(Rect)
    else
      if not Manager.Animated then begin // Do not change size of visible window before animation
        Manager.AnimWindow.Width := w;
        Manager.AnimWindow.Height := h;
      end;

    if acLayered then
      if Manager.Animated then begin
        FreeAndNil(Manager.ShowTimer);
        Manager.ShowTimer := TsShowTimer.Create(nil);
        Manager.ShowTimer.Enabled := False;
        Manager.ShowTimer.Init(Self, w, h)
      end
      else begin
        if IsNTFamily then begin
          CreateAlphaBmp(w, h);
          SetFormBlendValue(Manager.AnimWindow.Handle, AlphaBmp, MaxByte);
          ShowWindow(Manager.AnimWindow.Handle, SW_SHOWNOACTIVATE);
        end;
        SetWindowPos(Manager.AnimWindow.Handle, HWND_TOPMOST, Manager.NewBounds.Left, Manager.NewBounds.Top, w, h, SWP_SHOWWINDOW or SWP_NOACTIVATE or SWP_NOOWNERZORDER);

        if AlphaBmp <> nil then begin
          if Manager.OldAlphaBmp <> nil then
            Manager.OldAlphaBmp.Free;

          Manager.OldAlphaBmp := AlphaBmp;
          AlphaBmp := nil;
        end;
      end;

    Manager.State := hsNormal;
    Manager.FHintPos.x := -1;
{$ELSE}
    inherited;
{$ENDIF}
    if Manager.TmpShowData.Control <> nil then
      Manager.HintTimeCtrl := Manager.TmpShowData.Control
    else
      Manager.HintTimeCtrl := CtrlUnderMouse;

    Manager.StartHideTimer;
  end;
end;


constructor TacCustomHintWindow.Create(AOwner: TComponent);
begin
  inherited;
  BorderWidth := 0;
  if not (Self is TacTrackHintWindow) then
    acHintWindow := Self;
end;


procedure TacCustomHintWindow.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  Params.Style := Params.Style and not WS_BORDER;
end;


{$IFDEF DEVEX2011}
procedure TacCustomHintWindow.CreateWnd;
begin
  inherited;
  if ParentWindow = 0 then
    ParentWindow := Application.Handle;
end;
{$ENDIF}


function TacCustomHintWindow.CalcHintRect(MaxWidth: Integer; const AHint: string; AData: Pointer): TRect;
var
  sHTML: TsHtml;
  Lnks: TacLinks;
  i, w, h: integer;
  HintText: acString;
  HintInfo: THintInfo;
begin
  if Assigned(Manager) then begin
    SkinIndex := Manager.GetSkinIndex;
    if Manager.TmpShowData.Width <> 0 then
      Result := Rect(0, 0, Manager.TmpShowData.Width, Manager.TmpShowData.Height)
    else begin
      if Manager.HintFrame <> nil then
        Result := MkRect(Manager.HintFrame)
      else begin
{$IFDEF TNTUNICODE}
        if (TntApplication <> nil) and (TntApplication.Hint <> '') then
          HintText := TntApplication.Hint
        else
{$ENDIF}
          HintText := AHint;

//        HandleNeeded; // Handle required for PPI receiving ?
        Result := MkRect(iff(Manager.MaxWidth <= 0, Screen.Width, Manager.MaxWidth * sSkinManager.CurrentPPI(nil, Self) div 96), 0);
{$IFNDEF ACHINTS}
        if Manager.CustomFont <> nil then
          Manager.FCacheBmp.Canvas.Font.Assign(Manager.CustomFont)
        else begin
          if Manager.Skinned then
            Manager.FCacheBmp.Canvas.Font.Assign(Screen.HintFont)
          else
            Manager.FCacheBmp.Canvas.Font.Assign(Template.FFont);

          UpdateFontSize(Manager.FCacheBmp.Canvas.Font, not Manager.Skinned);
        end;
{$ENDIF}
        if Manager.FHTMLMode then begin
          sHTML := TsHtml.Create;
          sHTML.Init(Manager.FCacheBmp, HintText, Result);
          Result := sHTML.HtmlText(Lnks, True);
          FreeAndNil(sHTML);
        end
        else
          acDrawText(Manager.FCacheBmp.Canvas.Handle, PacChar(HintText), Result, DT_CALCRECT or acAlignmentArray[Manager.TextAlignment] or DT_VCENTER or DT_WORDBREAK or DT_NOPREFIX or DrawTextBiDiModeFlagsReadingOnly);
      end;

      if Assigned(Manager.OnMeasure) then begin
        HintInfo.HintControl := Manager.CurrentHintInfo.HintControl;
        HintInfo.HintPos := Manager.CurrentHintInfo.HintPos;
        w := WidthOf(Result);
        h := HeightOf(Result);
        Manager.OnMeasure(HintText, HintInfo, w, h);
        Result.Right := Result.Left + w;
        Result.Bottom := Result.Top + h;
      end;
    end;

    if Manager.Skinned then begin
      i := SkinMargin(0) + SkinMargin(2) + SkinBorderWidth * 2;
      Inc(Result.Right, i);
      Inc(Result.Bottom, SkinMargin(1) + SkinMargin(3) + SkinBorderWidth * 2);
    end;
  end;
{$IFDEF DEVEXHINTS}
  ParentWindow := Application.Handle; // Workaround for dxHints, prevent of wnd recreation
{$ENDIF}
end;


procedure TacCustomHintWindow.WMEraseBkGND(var Message: TWMPaint);
begin
  Message.Result := 1;
end;


procedure TacCustomHintWindow.WMNCPaint(var Message: TWMPaint);
begin
  if Assigned(Manager) then
    PrepareMask;

  Message.Result := 1;
end;


procedure TacCustomHintWindow.TextOut(Bmp: TBitmap);
const
  TextIdent = 4;
var
  DefaultDraw: boolean;
  aText: acString;
  Lnks: TacLinks;
{$IFNDEF ACHINTS}
  Flags: Integer;
{$ENDIF}
  R, RCalc, RText: TRect;
  GlyphPos: TPoint;
  sd: TsCommonData;
  SaveIndex: hdc;
  sHTML: TsHtml;
  i: integer;
begin
  R := MainRect;
  if Manager.Skinned then begin
    R.Left   := R.Left   + SkinMargin(0) + SkinBorderWidth;
    R.Top    := R.Top    + SkinMargin(1) + SkinBorderWidth;
    R.Right  := R.Right  - SkinMargin(2) - SkinBorderWidth;
    R.Bottom := R.Bottom - SkinMargin(3) - SkinBorderWidth;
  end
  else begin
    inc(R.Left,   Image.ClientMargins.Left);
    inc(R.Top,    Image.ClientMargins.Top);
    dec(R.Right,  Image.ClientMargins.Right);
    dec(R.Bottom, Image.ClientMargins.Bottom);
  end;

  DefaultDraw := True;
  if Manager.TmpShowData.Width <> 0 then begin // If manual showing with HintShowData struct
    if Manager.TmpShowData.Image <> nil then begin
      case Manager.TmpShowData.ImageAlignment of
        asLeft: begin
          GlyphPos.X := R.Left;
          GlyphPos.Y := R.Top + (Manager.TmpShowData.Height - Manager.TmpShowData.Image.Height) div 2;
          inc(R.Left, Manager.TmpShowData.Image.Width + TextIdent);
        end;

        asRight: begin
          GlyphPos.X := R.Left + Manager.TmpShowData.Width - Manager.TmpShowData.Image.Width;
          GlyphPos.Y := R.Top + (Manager.TmpShowData.Height - Manager.TmpShowData.Image.Height) div 2;
          dec(R.Right, Manager.TmpShowData.Image.Width + TextIdent);
        end;

        asTop: begin
          GlyphPos.X := R.Left + (Manager.TmpShowData.Width - Manager.TmpShowData.Image.Width) div 2;
          GlyphPos.Y := R.Top;
          inc(R.Top, Manager.TmpShowData.Image.Height + TextIdent);
        end;

        asBottom: begin
          GlyphPos.X := R.Left + (Manager.TmpShowData.Width - Manager.TmpShowData.Image.Width) div 2;
          GlyphPos.Y := R.Top + Manager.TmpShowData.Height - Manager.TmpShowData.Image.Height;
          dec(R.Bottom, Manager.TmpShowData.Image.Height + TextIdent);
        end;
      end;

      if Manager.TmpShowData.Image.PixelFormat = pf32bit then
        CopyBmp32(Rect(GlyphPos.X, GlyphPos.Y, GlyphPos.X + Manager.TmpShowData.Image.Width, GlyphPos.Y + Manager.TmpShowData.Image.Height),
                MkRect(Manager.TmpShowData.Image), Bmp, Manager.TmpShowData.Image, EmptyCI, False, Graphics.clNone, 0, False)
      else
        Bmp.Canvas.Draw(GlyphPos.X, GlyphPos.Y, Manager.TmpShowData.Image);
    end;

    if Manager.TmpShowData.Caption <> '' then begin
      RCalc := MkRect;
      RCalc.Right := WidthOf(R);
      if Manager.FHTMLMode then begin
        sHTML := TsHtml.Create;
        sHTML.Init(Bmp, Manager.TmpShowData.Caption, RCalc);
        RCalc := sHTML.HtmlText(Lnks, True);
        FreeAndNil(sHTML);
      end
      else
        acDrawText(Bmp.Canvas.Handle, PacChar(Manager.TmpShowData.Caption), RCalc, DT_NOPREFIX or DT_WORDBREAK or DT_CALCRECT);

      RText.Left := R.Left + (WidthOf(R) - WidthOf(RCalc)) div 2;
      RText.Top := R.Top + (HeightOf(R) - HeightOf(RCalc)) div 2;
      RText.Right := RText.Left + WidthOf(RCalc);
      RText.Bottom := RText.Top + HeightOf(RCalc);
    end;

    if Manager.Skinned then begin
      if Manager.FHTMLMode then begin
        sHTML := TsHtml.Create;
        sHTML.AlphaChannel := True;
        sHTML.BidiMode := BidiMode;
        sHTML.Alignment := Manager.TextAlignment;
        sHTML.Init(Bmp, Manager.TmpShowData.Caption, rText);
        sHTML.HtmlText(Lnks);
        FreeAndNil(sHTML);
      end
      else begin
        Flags := DT_EXPANDTABS or DT_NOCLIP or DT_WORDBREAK or acAlignmentArray[Manager.TextAlignment] or DrawTextBiDiModeFlagsReadingOnly or DT_VCENTER;
        if Application.IsRightToLeft then
          Flags := Flags or DT_RTLREADING;

        WriteText32(Bmp, PacChar(Manager.TmpShowData.Caption), True, RText, Flags, SkinIndex, 0, DefaultManager);
      end;
    end
    else begin
      if Template <> nil then
        Bmp.Canvas.Font.Color := Template.Font.Color;

      Bmp.Canvas.Brush.Style := bsClear;
      if Bmp.Canvas.Font.Color < 0 then
        Bmp.Canvas.Font.Color := 0;

      if Manager.FHTMLMode then begin
        sHTML := TsHtml.Create;
        sHTML.AlphaChannel := True;
        sHTML.BidiMode := BidiMode;
        sHTML.Alignment := Manager.TextAlignment;
        sHTML.Init(Bmp, Manager.TmpShowData.Caption, rText);
        sHTML.HtmlText(Lnks);
        FreeAndNil(sHTML);
      end
      else
        acDrawText(Bmp.Canvas.Handle, PacChar(Manager.TmpShowData.Caption), RText, acAlignmentArray[Manager.TextAlignment] or DT_NOPREFIX or DT_WORDBREAK or DrawTextBiDiModeFlagsReadingOnly or DT_VCENTER);
    end;
  end

  else begin
    if Assigned(Manager.OnPaint) then begin
{$IFDEF TNTUNICODE}
      if (TntApplication <> nil) and (TntApplication.Hint <> '') then
        aText := TntApplication.Hint
      else
{$ENDIF}
        aText := iff(Text <> '', Text, Caption);

      Manager.OnPaint(aText, Manager.CurrentHintInfo, R, Bmp, DefaultDraw);
    end;

    if DefaultDraw then
      if Manager.HintFrame <> nil then begin
        Manager.HintFrame.SetBounds(R.Left, R.Top, Manager.HintFrame.Width, Manager.HintFrame.Height);
{       // Workaround for dxHints, prevent of wnd recreation
        Manager.HintFrame.Visible := False;
        Manager.HintFrame.SetBounds(R.Left, R.Top, Manager.HintFrame.Width, Manager.HintFrame.Height);
        Manager.HintFrame.Parent := Self;
        Manager.HintFrame.Visible := True;
        Manager.HintFrame.HandleNeeded;
}
        if Manager.Skinned then
          Manager.HintFrame.Font.Color := DefaultManager.CommonSkinData.gd[SkinIndex].Props[0].Fontcolor.Color
        else
          if (Template <> nil) then
            Manager.HintFrame.Font.Color := Template.Font.Color;

        sd := GetCommonData(Manager.HintFrame.Handle);
        if sd = nil then
          sd := TsCommonData(Manager.HintFrame.Perform(SM_ALPHACMD, AC_GETSKINDATA_HI, 0));

        if (sd <> nil) and Assigned(sd.WndProc) then begin
          OldSDProc := sd.WndProc;
          sd.WndProc := FrameWndProc;
        end;
        OldFrameProc := Manager.HintFrame.WindowProc;
        Manager.HintFrame.WindowProc := FrameWndProc;

        SaveIndex := SaveDC(Bmp.Canvas.Handle);
        Bmp.Canvas.Lock;
        for i := 0 to Manager.HintFrame.ControlCount - 1 do
          if Manager.HintFrame.Controls[i].Visible then begin
            MovewindowOrg(Bmp.Canvas.Handle, R.Left + Manager.HintFrame.Controls[i].Left, R.Top + Manager.HintFrame.Controls[i].Top);
            Manager.HintFrame.Controls[i].Perform(WM_PAINT, WPARAM(Bmp.Canvas.Handle), 0);
            MovewindowOrg(Bmp.Canvas.Handle, -R.Left - Manager.HintFrame.Controls[i].Left, -R.Top - Manager.HintFrame.Controls[i].Top);
            FillAlphaRect(Bmp, R, MaxByte);
          end;

        if (sd <> nil) and Assigned(sd.WndProc) then begin
          sd.WndProc := OldSDProc;
          OldSDProc := nil;
        end;
        Manager.HintFrame.WindowProc := OldFrameProc;
        OldFrameProc := nil;
        // Next two line are called after labels painting for avoiding of "Reflected text" error
        Bmp.PixelFormat := pf32bit;
        Bmp.HandleType := bmDIB;
        Bmp.Canvas.UnLock;
        RestoreDC(Bmp.Canvas.Handle, SaveIndex);
      end
      else begin
//        OffsetRect(R, -1, 0);
        Bmp.Canvas.Brush.Style := bsClear;
        Bmp.Canvas.Pen.Style := psSolid;
{$IFDEF TNTUNICODE}
        if (TntApplication <> nil) and (TntApplication.Hint <> '') then
          aText := TntApplication.Hint
        else
{$ENDIF}
          aText := iff(Text <> '', Text, Caption);

{$IFNDEF ACHINTS}
        if Manager.CustomFont <> nil then
          Bmp.Canvas.Font.Assign(Manager.CustomFont)
        else begin
          if Manager.Skinned then
            Bmp.Canvas.Font.Assign(Screen.HintFont)
          else
            Bmp.Canvas.Font.Assign(Template.FFont);

          UpdateFontSize(Bmp.Canvas.Font, not Manager.Skinned);
        end;
{$ENDIF}
        if Manager.FHTMLMode then begin
{$IFNDEF ACHINTS}
          if Manager.CustomFont = nil then begin
            if Manager.Skinned then
              Bmp.Canvas.Font.Color := DefaultManager.CommonSkinData.gd[SkinIndex].Props[0].Fontcolor.Color
            else
              if (Template <> nil) then
                Bmp.Canvas.Font.Color := Template.Font.Color;

            if Bmp.Canvas.Font.Color < 0 then
              Bmp.Canvas.Font.Color := 0;
          end;
{$ENDIF}
          sHTML := TsHtml.Create;
          sHTML.AlphaChannel := True;
          sHTML.BidiMode := BidiMode;
          sHTML.Alignment := Manager.TextAlignment;
          sHTML.Init(Bmp, aText, R);
          sHTML.HtmlText(Lnks);
          FreeAndNil(sHTML);
        end
        else
{$IFNDEF ACHINTS}
          if Manager.Skinned then begin
            Flags := DT_EXPANDTABS or DT_NOCLIP or DT_WORDBREAK or acAlignmentArray[Manager.TextAlignment] or DrawTextBiDiModeFlagsReadingOnly;
            if Application.IsRightToLeft then
              Flags := Flags or DT_RTLREADING;

            if Manager.CustomFont = nil then
              WriteText32(Bmp, PacChar(aText), True, R, Flags, SkinIndex, 0, DefaultManager)
            else
              WriteText32(Bmp, PacChar(aText), True, R, Flags, -1, 0, DefaultManager);
          end
          else
{$ENDIF}
          begin
            if Manager.CustomFont = nil then begin
              if (Template <> nil) then
                Bmp.Canvas.Font.Color := Template.Font.Color;

              if Bmp.Canvas.Font.Color < 0 then
                Bmp.Canvas.Font.Color := 0;
            end;
            acDrawText(Bmp.Canvas.Handle, PacChar(aText), R, acAlignmentArray[Manager.TextAlignment] or DT_NOPREFIX or DT_WORDBREAK or DrawTextBiDiModeFlagsReadingOnly);
          end;
      end;
  end;
end;


function TacCustomHintWindow.MainRect: TRect;
begin
  Result := MkRect(WidthOf(Manager.NewBounds), HeightOf(Manager.NewBounds));
end;


procedure TacCustomHintWindow.WndProc(var Message: TMessage);
begin
{$IFDEF LOGGED}
  AddToLog(Message);
{$ENDIF}
  if Message.Msg = SM_ALPHACMD then
    case Message.WParamHi of
      AC_CTRLHANDLED: begin
        Message.Result := 1;
        Exit
      end; // AlphaSkins supported

      AC_GETBG: begin
        PacBGInfo(Message.LParam)^.Bmp := BodyBmp;
        PacBGInfo(Message.LParam)^.Offset := MkPoint;
        if BodyBmp <> nil then
          PacBGInfo(Message.LParam)^.BgType := btCache
        else begin
          PacBGInfo(Message.LParam)^.BgType := btFill;
          PacBGInfo(Message.LParam)^.Color := Color;
        end;
        Exit;
      end;

      AC_CHILDCHANGED:
        Message.LParam := 1;
    end;

  inherited;
  if Manager <> nil then
    case Message.Msg of
      WM_SHOWWINDOW:
        case Message.WParam of
          1: // Show
            acHintWindow := Self;

          0: // Hide
            OnCloseHint;
        end;

      WM_SIZE, WM_MOVE:
        if (Manager.AnimWindow <> nil) and (Manager.State <> hsInit) then begin
          Manager.AnimWindow.SetBounds(Left, Top, Manager.AnimWindow.Width, Manager.AnimWindow.Height);
          Manager.NewBounds := Manager.AnimWindow.BoundsRect;
        end;
    end;
end;


procedure TacCustomHintWindow.FrameWndProc(var Message: TMessage);
begin
  if Message.Msg = SM_ALPHACMD then
    case Message.WParamHi of
      AC_GETBG: begin
        PacBGInfo(Message.LParam)^.Bmp := AlphaBmp;
        PacBGInfo(Message.LParam)^.Offset := MkPoint;
        if AlphaBmp <> nil then
          PacBGInfo(Message.LParam)^.BgType := btCache
        else begin
          PacBGInfo(Message.LParam)^.BgType := btFill;
          PacBGInfo(Message.LParam)^.Color := Color;
        end;
        Exit;
      end
      else
        OldFrameProc(Message);
    end
  else
    OldFrameProc(Message);
end;


procedure TacCustomHintWindow.DoDestroy;
begin
  OnCloseHint;
  if Manager <> nil then begin
{$IFDEF D2006}
    if not (csRecreating in ControlState) then
{$ENDIF}
      Manager.ResetHintInfo;
      
    if AlphaBmp <> nil then begin
      if Manager.OldAlphaBmp <> nil then
        FreeAndNil(Manager.OldAlphaBmp);

      Manager.OldAlphaBmp := AlphaBmp;
      AlphaBmp := nil;
    end;
    if Manager.HintWindow = Self then
      Manager.HintWindow := nil;
  end
  else
    if Assigned(AlphaBmp) then
      FreeAndNil(AlphaBmp);

  if Assigned(MaskBmp) then
    FreeAndNil(MaskBmp);

  if Assigned(BodyBmp) then
    FreeAndNil(BodyBmp);
end;


destructor TacCustomHintWindow.Destroy;
begin
  DoDestroy;
  inherited;
end;


procedure TacCustomHintWindow.WMDestroy(var Message: TWMDestroy);
begin
  DoDestroy;
  inherited;
end;


function TacCustomHintWindow.GetMousePosition: TPoint;
begin
  if (Manager.FHintPos.x = -1) then
    Result := acMousePos
  else
    Result := Manager.FHintPos;
end;


procedure TacCustomHintWindow.Paint;
begin
//
end;


procedure TacCustomHintWindow.PrepareMask;
begin
  rgn := 0;
  FreeAndNil(MaskBmp);
  if not AlphaUsed then begin
    MaskBmp := GetMask;
{$IFNDEF ACHINTS}
    if Assigned(MaskBmp) and Manager.Skinned and not (Self is TacTrackHintWindow) then begin // Defining window region by MaskBmp
try
      GetRgnFromBmp(rgn, MaskBmp, clWhite);
      SetWindowRgn(Handle, rgn, False);
finally

end;
    end
{$ENDIF}
  end;
end;


function TacCustomHintWindow.GetMask: TBitmap;
{$IFNDEF ACHINTS}
var
  White, Black: TsColor_;
  x, y, h, w: integer;
  S0, S: PRGBAArray_;
  DeltaS: integer;
  CI: TCacheInfo;
  CC: TsColor;
{$ENDIF}
begin
{$IFNDEF ACHINTS}
  CC.I := -65282;   // R:254 G:0 B:255 A:255
  CI.FillColor := CC.C;
  CI.Ready := False;
  CC.A := 0;
  CI.Bmp := nil;
  White.C := clWhite;
  Black.I := 0;
  if Assigned(AlphaBmp) then
    Result := CreateBmpLike(AlphaBmp)
  else
    Result := CreateBmpLike(Manager.FCacheBmp);

  if DefaultManager.IsActive then
    PaintItem(SkinIndex, CI, True, 0, MkRect(Self), MkPoint, Result, DefaultManager.CommonSkinData)
  else
    FillDC(Result.Canvas.Handle, MkRect(Result), 0);

  h := Result.Height - 1;
  w := Result.Width - 1;
  CC.B := 254;
  CC.R := MaxByte;
  CC.A := 255;
  if InitLine(Result, Pointer(S0), DeltaS) then
    for y := 0 to h do begin
      S := Pointer(PAnsiChar(S0) + DeltaS * Y);
      for x := 0 to w do
        with S[X] do
          if C = CC.C then
            C := $FFFFFF
          else
            C := 0;
    end;
{$ENDIF}
end;


procedure PaintHint(SkinIndex: integer; aRect: TRect; ShadowCalculated, AlphaUsed: boolean; Result: TBitmap);
var
  Mode: integer;
  CI: TCacheInfo;
  ABmp, SBmp, ShadowLayer: TBitmap;
  MaxWidths: TRect;
begin
  CI.Ready := False;
  CI.FillColor := sFuchsia.C;
  TsColor(CI.FillColor).A := 0;
  with DefaultManager.CommonSkinData, gd[SkinIndex] do
    if Manager.ShadowCalculated then begin
      ShadowLayer := CreateBmpLike(Result);
      FillRect32(ShadowLayer, MkRect(ShadowLayer), 0, 0);

      MaxWidths := OuterOffset;
      if gd[SkinIndex].BorderIndex >= 0 then begin
        inc(MaxWidths.Left,   ma[gd[SkinIndex].BorderIndex].WL);
        inc(MaxWidths.Top,    ma[gd[SkinIndex].BorderIndex].WT);
        inc(MaxWidths.Right,  ma[gd[SkinIndex].BorderIndex].WR);
        inc(MaxWidths.Bottom, ma[gd[SkinIndex].BorderIndex].WB);
      end;

      DrawSkinRect32Ex(ShadowLayer, MkRect(ShadowLayer), EmptyCI, ma[OuterMask], 0, MaxWidths, OuterOpacity);

      aRect.Left := max(0, aRect.Left + OuterOffset.Left);
      aRect.Top  := max(0, aRect.Top + OuterOffset.Top);
      aRect.Right  := min(Result.Width, aRect.Right - OuterOffset.Right);
      aRect.Bottom := min(Result.Height, aRect.Bottom - OuterOffset.Bottom);

      CopyBmp(Result, ShadowLayer);
      CI := MakeCacheInfo(ShadowLayer);
      PaintItem(SkinIndex, CI, True, 0, aRect, MkPoint, Result, DefaultManager.CommonSkinData);
      ShadowLayer.Free;
    end
    else
      if (gd[SkinIndex].Props[0].Transparency = 100) and (BorderIndex >= 0) then
        with ma[BorderIndex] do begin // Used BorderMask
          if Bmp = nil then begin
            SBmp := DefaultManager.MasterBitmap;
            aRect := R;
          end
          else begin
            SBmp := Bmp;
            aRect := MkRect(Bmp);
          end;

          if not AlphaUsed then begin
            ABmp := sGraphUtils.CreateAlphaBmp(SBmp, aRect);
            aRect := MkRect(ABmp);
          end
          else
            ABmp := SBmp;

          Mode := integer(ma[BorderIndex].DrawMode and BDM_STRETCH = BDM_STRETCH);
          PaintControlByTemplate(Result, ABmp, MkRect(Result), aRect,
                                 Rect(WL, WT, WR, WB), Rect(WL, WT, WR, WB),
                                 Rect(Mode, Mode, Mode, Mode), True, True); // For internal shadows - stretch only allowed
          if not AlphaUsed then
            ABmp.Free;

          UpdateAlpha(Result);
        end
      else
        PaintItem(SkinIndex, CI, True, 0, MkRect(Result), MkPoint, Result, DefaultManager.CommonSkinData);
end;


function TacCustomHintWindow.GetBody: TBitmap;
{$IFNDEF ACHINTS}
var
  aRect: TRect;
{$ENDIF}
begin
{$IFNDEF ACHINTS}
  if AlphaBmp <> nil then
    Result := CreateBmpLike(AlphaBmp)
  else
    Result := CreateBmpLike(Manager.FCacheBmp);

  if Manager.HintFrame <> nil then
    BodyBmp := Result;

  aRect := MkRect(Result);

  PaintHint(SkinIndex, aRect, Manager.ShadowCalculated, AlphaUsed, Result);
  TextOut(Result);
{$ENDIF}
end;


function TacCustomHintWindow.SkinMargin(Border: byte): integer;
begin
{$IFNDEF ACHINTS}
  with DefaultManager.CommonSkinData, gd[SkinIndex] do begin
    if BorderIndex >= 0 then
      with ma[BorderIndex] do
        case Border of
          0:   Result := WL;
          1:   Result := WT;
          2:   Result := WR
          else Result := WB;
        end
    else
{$ENDIF}
      Result := 0;

    if Manager.ShadowCalculated then
      case Border of
        0:   inc(Result, OuterOffset.Left);
        1:   inc(Result, OuterOffset.Top);
        2:   inc(Result, OuterOffset.Right);
        else inc(Result, OuterOffset.Bottom);
      end;
  end;
end;


procedure TacCustomHintWindow.CreateAlphaBmp;
var
  c: TsColor_;
  D0, S0, M0, SH0, D, S, M, SH: PRGBAArray_;
  x, y, DeltaD, DeltaS, DeltaM, DeltaSH: integer;
begin
  if AlphaBmp <> nil then
    FreeAndNil(AlphaBmp);

  AlphaBmp := CreateBmp32(Width, Height);
  with DefaultManager.CommonSkinData, gd[SkinIndex] do begin
    if BorderIndex >= 0 then
      AlphaUsed := ma[BorderIndex].DrawMode and BDM_ALPHA <> 0
    else
      AlphaUsed := False;

    PrepareMask;
    if Assigned(BodyBmp) then
      FreeAndNil(BodyBmp);

    BodyBmp := GetBody;
    if (Props[0].Transparency = 100) and (BorderIndex >= 0) or Manager.ShadowCalculated then { BorderMask used }
      AlphaBmp.Assign(BodyBmp)
    else begin
      FillDC(Manager.FCacheBmp.Canvas.Handle, MkRect(Self), clWhite);
      if InitLine(BodyBmp, Pointer(S0), DeltaS) and
           InitLine(AlphaBmp, Pointer(D0), DeltaD) and
             InitLine(MaskBmp, Pointer(M0), DeltaM) and
               InitLine(Manager.FCacheBmp, Pointer(SH0), DeltaSH) then
        for y := 0 to Height - 1 do begin
          S := Pointer(PAnsiChar(S0) + DeltaS * Y);
          D := Pointer(PAnsiChar(D0) + DeltaD * Y);
          M := Pointer(PAnsiChar(M0) + DeltaM * Y);
          SH := Pointer(PAnsiChar(SH0) + DeltaSH * Y);
          for x := 0 to Width - 1 do begin
            if M[X].R = MaxByte then begin
              c.I := 0;
              c.A := MaxByte - SH[X].R;
            end
            else begin
              c := S[X];
              if S[X].C = sFuchsia.C then
                c.A := 0
              else
                c.A := MaxByte;
            end;
            D[X] := c;
          end;
        end;
    end;
  end;
  if Assigned(BodyBmp) then
    FreeAndNil(BodyBmp);
end;


function TacCustomHintWindow.GetArrowPos(var Rect: TRect; const mPos: TPoint): TacMousePosition;
var
  t, l, Auto: boolean;
  h, w: integer;
{$IFDEF DELPHI6UP}
  Monitor: TMonitor;
{$ENDIF}

  procedure SetLeft(Value: integer; al: boolean);
  begin
    Rect.Left := Value;
    l := al;
    Auto := True;
  end;

  procedure SetTop(Value: integer; at: boolean);
  begin
    Rect.Top := Value;
    t := at;
    Auto := True;
  end;

begin
  FMousePos := Manager.FDefaultMousePos;
  Result := FMousePos;

  dec(Rect.Right,  ShadowSizes.Right);
  dec(Rect.Bottom, ShadowSizes.Bottom);
  inc(Rect.Left,   ShadowSizes.Left);
  inc(Rect.Top,    ShadowSizes.Top);

  w := WidthOf(Rect);
  h := HeightOf(Rect);
  t := not (FMousePos in [mpLeftBottom, mpRightBottom]);
  l := not (FMousePos in [mpRightTop, mpRightBottom]);
  if not t then
    OffsetRect(Rect, 0, - h);

  if not l then
    OffsetRect(Rect, - w, 0);

  Auto := False; // Calc arrow position
{$IFDEF DELPHI6UP}
  Monitor := Screen.MonitorFromPoint(Rect.TopLeft);
  if Monitor = nil then
    Exit;
    
  if Rect.Bottom > Monitor.Top + Monitor.Height then
    if Manager.Skinned then
      SetTop(Monitor.Top + Monitor.Height - h, False)
    else
      if Template.SingleImageUsed then
        SetTop(mPos.Y - (Rect.Bottom - (Monitor.Top + Monitor.Height)), False)
      else
        SetTop(mPos.Y - h, False);

  if Rect.Top < Monitor.Top then
    SetTop(Monitor.Top, True);

  if Rect.Right > Monitor.Left + Monitor.Width then
    if Manager.Skinned then
      SetLeft(Monitor.Left + Monitor.Width - w, False)
    else
      if Template.SingleImageUsed then
        SetLeft(mPos.X - (Rect.Right - (Monitor.Left + Monitor.Width)), False)
      else
        SetLeft(mPos.X - w, False);

  if Rect.Left < Monitor.Left then
    SetLeft(Monitor.Left, True);
{$ELSE}
  if Rect.Bottom > Screen.Height then
    SetTop(mPos.y - h, False);

  if Rect.Top < 0 then
    SetTop(mPos.y, True);

  if Rect.Right > Screen.Width then
    SetLeft(mPos.x - w, False);

  if Rect.Left < 0 then
    SetLeft(mPos.x, True);
{$ENDIF}
  if Auto then
    Result := pArray[t, l];

  inc(Rect.Right,  ShadowSizes.Right);
  inc(Rect.Bottom, ShadowSizes.Bottom);
  dec(Rect.Left,   ShadowSizes.Left);
  dec(Rect.Top,    ShadowSizes.Top);
end;


procedure TacCustomHintWindow.UpdateWnd;
var
  lTicks: DWord;
  i, StepCount: integer;
  FBlend: TBlendFunction;
  SrcRect, DstRect: TRect;
  StepB, Blend, dx, dy, x, y: real;
begin
  if w = 0 then
    if acHintWindow <> nil then
      w := acHintWindow.Width
    else
      w := Manager.AnimWindow.Width;

  if h = 0 then
    if acHintWindow <> nil then
      h := acHintWindow.Height
    else
      h := Manager.AnimWindow.Height;

  CreateAlphaBmp(w, h);
  if not Manager.HintShowing then
    SetFormBlendValue(Manager.AnimWindow.Handle, AlphaBmp, 0);

  FBlend := DefBlend;
  // Show window with hint
  if Manager.Animated and IsNTFamily and not Manager.HintShowing then begin
    if not (csDestroying in ComponentState) then
      with DefaultManager.CommonSkinData.gd[SkinIndex] do begin
        if Manager.Skinned and (Props[0].Transparency <> 100) then
          i := Props[0].Transparency
        else
          i := 0;

        StepCount := max(DefAnimationTime div acTimerInterval, 1);
        FBlend.SourceConstantAlpha := 0;
        if Manager.AnimWindow <> nil then begin
          SrcRect := Manager.AnimWindow.BoundsRect;
          DstRect := Manager.NewBounds;
          FBlend.SourceConstantAlpha := MaxByte - i;
          dx := (DstRect.Left - SrcRect.Left) / StepCount;
          dy := (DstRect.Top - SrcRect.Top) / StepCount;
          x := SrcRect.Left;
          y := SrcRect.Top;
          for i := 0 to StepCount - 1 do begin
            x := x + dx;
            y := y + dy;
            lTicks := GetTickCount;
            Manager.AnimWindow.Left := Round(x);
            Manager.AnimWindow.Top  := Round(y);
            WaitTicks(lTicks);
          end;
          if (Manager.NewWindow <> nil) then
            FreeAndNil(Manager.NewWindow);

          SetFormBlendValue(Manager.AnimWindow.Handle, AlphaBmp, MaxByte);
          ShowWindow(Manager.AnimWindow.Handle, SW_SHOWNOACTIVATE);
        end
        else begin
          i := Max(0, Min(100, i));
          StepB := Round(MaxByte - 2.55 * i) / StepCount;
          Blend := 0;
          SetFormBlendValue(Manager.AnimWindow.Handle, AlphaBmp, 0);
          ShowWindow(Manager.AnimWindow.Handle, SW_SHOWNOACTIVATE);
          for i := 0 to StepCount - 1 do begin
            lTicks := GetTickCount;
            Blend := Blend + StepB;
            if not (csDestroying in ComponentState) then
              SetFormBlendValue(Manager.AnimWindow.Handle, AlphaBmp, Round(Blend))
            else
              Break;

            WaitTicks(lTicks);
          end;
        end;
      end;
  end
  else
    if IsNTFamily then begin
      SetFormBlendValue(Manager.AnimWindow.Handle, AlphaBmp, MaxByte);
      SetWindowPos(Manager.AnimWindow.Handle, HWND_TOPMOST, Manager.NewBounds.Left, Manager.NewBounds.Top, w, h, SWP_SHOWWINDOW or SWP_NOACTIVATE or SWP_NOOWNERZORDER);// or SWP_NOCOPYBITS);
      ShowWindow(Manager.AnimWindow.Handle, SW_SHOWNOACTIVATE);
    end
    else
      SetWindowPos(Manager.AnimWindow.Handle, HWND_TOPMOST, 0, 0, w, h, SWP_SHOWWINDOW or SWP_NOACTIVATE or SWP_NOMOVE or SWP_NOOWNERZORDER);// or SWP_NOCOPYBITS);

  Manager.HintShowing := True;
  if AlphaBmp <> nil then begin
    if Manager.OldAlphaBmp <> nil then
      Manager.OldAlphaBmp.Free;

    Manager.OldAlphaBmp := AlphaBmp;
    AlphaBmp := nil;
  end;
end;


function TacPngHintWindow.CalcHintRect(MaxWidth: Integer; const AHint: string; AData: Pointer): TRect;
var
  DefRect: TRect;
  p: TPoint;

  procedure UpdateHintRect;
  begin
    Inc(Result.Right, Image.ClientMargins.Left + Image.ClientMargins.Right);
    Inc(Result.Bottom, Image.ClientMargins.Top + Image.ClientMargins.Bottom);
  end;

begin
  if Assigned(Manager) then begin
    if not Manager.Skinned then begin
      Manager.InitTemplate;
      if Manager.CustomFont <> nil then
        Manager.FCacheBmp.Canvas.Font.Assign(Manager.CustomFont)
      else begin
        Manager.FCacheBmp.Canvas.Font.Assign(Template.FFont);
        UpdateFontSize(Manager.FCacheBmp.Canvas.Font, True);
      end;

      DefRect := inherited CalcHintRect(MaxWidth, AHint, AData);

      Result := DefRect;
      UpdateHintRect;
      if Manager.TmpShowData.Control <> nil then begin
        Manager.FHintLocation := Manager.TmpShowData.Position;
        Manager.CurrentHintInfo.HintControl := Manager.TmpShowData.Control;
        Manager.CheckAutoHintPos;
        OffsetRect(Result, Manager.FHintLocation.x, Manager.FHintLocation.y);
      end
      else
        if Manager.TmpShowData.Width <> 0 then
          p := Manager.TmpShowData.Position
        else begin
          if (Manager.FHintLocation.X = 0) or (Manager.FHintLocation.Y = 0) then
            p := GetMousePosition
          else
            p := Manager.FHintLocation;

          OffsetRect(Result, p.x, p.y);
          FMousePos := GetArrowPos(Result, p);
        end;

      Result := DefRect;
      UpdateImageTemplate;
      UpdateHintRect;
    end
    else begin
      DefRect := inherited CalcHintRect(MaxWidth, AHint, AData);
      if (Manager.TmpShowData.Control <> nil) then begin
        Manager.FHintLocation := Manager.TmpShowData.Position;
        Manager.CurrentHintInfo.HintControl := Manager.TmpShowData.Control;
        Manager.InitTemplate;
        Manager.CheckAutoHintPos;
        OffsetRect(Result, Manager.FHintLocation.x, Manager.FHintLocation.y);
      end;
      Result := DefRect;
    end
  end;
end;


function TacCustomHintWindow.ShadowSizes: TRect;
const
  Offset = 1;
begin
  if Manager.Active then
    if Manager.Skinned then
      if FullLayer then
        Result := Rect(max(0, SkinMargin(0) - Offset), max(0, SkinMargin(1) - Offset), max(0, SkinMargin(2) - Offset), max(0, SkinMargin(3) - Offset))
      else
        Result := MkRect
    else
      case FMousePos of
        mpLeftTop:     with Template.FImageDefault.ShadowSizes     do Result := Rect(Left, Top, Right, Bottom);
        mpRightTop:    with Template.FImageRightTop.ShadowSizes    do Result := Rect(Left, Top, Right, Bottom);
        mpLeftBottom:  with Template.FImageLeftBottom.ShadowSizes  do Result := Rect(Left, Top, Right, Bottom);
        mpRightBottom: with Template.FImageRightBottom.ShadowSizes do Result := Rect(Left, Top, Right, Bottom);
      end
  else
    Result := MkRect;
end;


constructor TacPngHintWindow.Create(AOwner: TComponent);
begin
  inherited;
end;


procedure TacPngHintWindow.CreateAlphaBmp(const Width, Height: integer);
var
  SrcBmp, TempBmp: TBitmap;
  FBlend: TBlendFunction;
begin
  if Manager.Skinned then
    inherited
  else begin
    if not (Self is TacTrackHintWindow) and HandleAllocated then
      SetWindowRgn(Handle, 0, False);

    if AlphaBmp <> nil then
      AlphaBmp.Free;

    AlphaBmp := CreateBmp32(Width, Height);
    InitBlendData(FBlend, MaxByte);
    if Template <> nil then begin
      if Image.FImage.Empty and (Image.FImageData.Size > 0) then begin
        Image.FImageData.Seek(0, 0);
        Image.FImage.LoadFromStream(Image.FImageData);
      end;
      if not Image.FImage.Empty then begin
        SrcBmp := Image.FImage;
        PaintControlByTemplate(AlphaBmp, SrcBmp, MkRect(AlphaBmp), MkRect(SrcBmp),
            Rect(Image.BordersWidths.Left, Image.BordersWidths.Top, Image.BordersWidths.Right, Image.BordersWidths.Bottom),
            Rect(MaxByte, MaxByte, MaxByte, MaxByte),
            Rect(ord(Image.BorderDrawModes.Left), ord(Image.BorderDrawModes.Top), ord(Image.BorderDrawModes.Right), ord(Image.BorderDrawModes.Bottom)),
            boolean(ord(Image.BorderDrawModes.Center)));
      end;
    end;
    TempBmp := TBitmap.Create;
    TempBmp.Assign(AlphaBmp);
    if not Manager.Skinned then begin
      Manager.InitTemplate;
      if FMousePos <> Manager.DefaultMousePos then
        UpdateImageTemplate;
    end;
    if Manager.CustomFont <> nil then
      AlphaBmp.Canvas.Font.Assign(Manager.CustomFont)
    else begin
      if Template <> nil then
        AlphaBmp.Canvas.Font.Assign(Template.FFont);

      UpdateFontSize(AlphaBmp.Canvas.Font, True);
    end;
    TextOut(AlphaBmp);
    CopyChannel32(AlphaBmp, TempBmp, 3);
    FreeAndNil(TempBmp);
    UpdateAlpha(AlphaBmp);
  end;
end;


procedure TacHintTemplate.AssignTo(Dest: TPersistent);
var
  aDst: TacHintTemplate;
begin
  if Dest <> nil then begin
    aDst := TacHintTemplate(Dest);
    ImageDefault   .AssignTo(aDst.ImageDefault);
    Img_LeftBottom .AssignTo(aDst.Img_LeftBottom);
    Img_RightBottom.AssignTo(aDst.Img_RightBottom);
    Img_RightTop   .AssignTo(aDst.Img_RightTop);
    aDst.HintAlign := HintAlign;
    aDst.Font.Assign(Font);
    aDst.Name := Name;
  end
  else
    inherited;
end;


constructor TacHintTemplate.Create(Collection: TCollection);
begin
  inherited Create(Collection);
  FOwner := TacHintTemplates(Collection);
  FImageDefault     := TacHintImage.Create(Self);
  FImageLeftBottom  := TacHintImage.Create(Self);
  FImageRightBottom := TacHintImage.Create(Self);
  FImageRightTop    := TacHintImage.Create(Self);
  FFont := TFont.Create;
  FFont.Height := -12;
  FHintAlign := haHorzCenter;
end;


destructor TacHintTemplate.Destroy;
begin
  FreeAndNil(FImageDefault);
  FreeAndNil(FImageLeftBottom);
  FreeAndNil(FImageRightBottom);
  FreeAndNil(FImageRightTop);
  FreeAndNil(FFont);
  inherited Destroy;
end;


function TacHintTemplate.SingleImageUsed: boolean;
begin
  Result := (Img_LeftBottom.FImageData.Size = 0) and ((Template = nil) or Template.FImageLeftBottom.FImage.Empty);
end;



procedure TacHintTemplate.SetFont(const Value: TFont);
begin
  FFont.Assign(Value);
  FOwner.FOwner.Changed;
end;


procedure TacHintTemplate.SetHintImage(const Index: Integer; const Value: TacHintImage);
begin
  case Index of
    0: FImageDefault    .Assign(Value);
    1: FImageLeftBottom .Assign(Value);
    2: FImageRightBottom.Assign(Value);
    3: FImageRightTop   .Assign(Value);
  end;
end;


procedure TacHintTemplates.AssignTo(Dest: TPersistent);
var
  i: integer;
  TmpSrc, TmpDst: TacHintTemplate;
begin
  if Dest <> nil then begin
    TacHintTemplates(Dest).Clear;
    for i := 0 to Count - 1 do begin
      TacHintTemplates(Dest).Add;
      TmpSrc := GetItem(i);
      TmpDst := TacHintTemplates(Dest).GetItem(i);
      TmpSrc.AssignTo(TmpDst);
    end;
  end
  else
    inherited;
end;


constructor TacHintTemplates.Create(AOwner: TsAlphaHints);
begin
  FOwner := AOwner;
  inherited Create(TacHintTemplate);
end;


destructor TacHintTemplates.Destroy;
begin
  inherited Destroy;
  FOwner := nil;
end;


function TacHintTemplates.GetItem(Index: Integer): TacHintTemplate;
begin
  Result := TacHintTemplate(inherited GetItem(Index));
end;


function TacHintTemplates.GetOwner: TPersistent;
begin
  Result := FOwner;
end;


procedure TacHintTemplates.SetItem(Index: Integer; Value: TacHintTemplate);
begin
  inherited SetItem(Index, Value);
end;


procedure TacBordersSizes.AssignTo(Dest: TPersistent);
begin
  if Dest <> nil then
    TacBordersSizes(Dest).SetValues(Left, Top, Right, Bottom)
  else
    inherited;
end;


constructor TacBordersSizes.Create(AOwner: TacHintImage);
begin
  FTop    := 0;
  FLeft   := 0;
  FBottom := 0;
  FRight  := 0;
  FOwner := AOwner;
end;


function TacBordersSizes.GetInteger(const Index: Integer): integer;
begin
  Result := 0;
  case Index of
    0: Result := FTop;
    1: Result := FLeft;
    2: Result := FBottom;
    3: Result := FRight;
  end;
  if Result < 0 then
    Result := 0;
end;


procedure TacBordersSizes.LoadFromStream(AStream: TStream);
begin
  AStream.Read(FTop,    SizeOf(FTop));
  AStream.Read(FLeft,   SizeOf(FLeft));
  AStream.Read(FRight,  SizeOf(FRight));
  AStream.Read(FBottom, SizeOf(FBottom));
end;


procedure TacBordersSizes.SaveToStream(AStream: TStream);
begin
  AStream.Write(FTop,    SizeOf(FTop));
  AStream.Write(FLeft,   SizeOf(FLeft));
  AStream.Write(FRight,  SizeOf(FRight));
  AStream.Write(FBottom, SizeOf(FBottom));
end;


procedure TacBordersSizes.SetValues(ALeft, ATop, ARight, ABottom: integer);
begin
  Left   := ALeft;
  Top    := ATop;
  Right  := ARight;
  Bottom := ABottom;
end;


procedure TacBordersSizes.SetInteger(Index, Value: integer);
var
  bChanged: boolean;

  procedure ChangeValue(var aProp: integer; aValue: integer);
  begin
    if aProp <> aValue then begin
      aProp := aValue;
      bChanged := True;
    end;
  end;

begin
  bChanged := False;
  case Index of
    0: ChangeValue(FTop,    Value);
    1: ChangeValue(FLeft,   Value);
    2: ChangeValue(FBottom, Value);
    3: ChangeValue(FRight,  Value);
  end;
  if bChanged and not acLocalUpd and (FOwner.FOwner.FOwner <> nil) then begin
    acLocalUpd := True;
    FOwner.FOwner.FOwner.FOwner.Changed;
    acLocalUpd := False;
  end;
end;


constructor TacHintImage.Create(AOwner: TacHintTemplate);
begin
  FImageData := TMemoryStream.Create;
  FImage := TPngGraphic.Create;
  FOwner := AOwner;
  FBordersWidths := TacBordersSizes.Create(Self);
  FClientMargins := TacBordersSizes.Create(Self);
  FShadowSizes   := TacBordersSizes.Create(Self);
  FBorderDrawModes := TacBorderDrawModes.Create(Self);
  FOffsetHorz := 0;
  FOffsetVert := 0;
end;


destructor TacHintImage.Destroy;
begin
  FreeAndNil(FBorderDrawModes);
  FreeAndNil(FClientMargins);
  FreeAndNil(FBordersWidths);
  FreeAndNil(FShadowSizes);
  FreeAndNil(FImage);
  FreeAndNil(FImageData);
  inherited Destroy;
end;


function TacHintImage.GetImgHeight: integer;
begin
  if FImage.Empty then
    Result := 0
  else
    Result := Image.Height;
end;


function TacHintImage.GetImgWidth: integer;
begin
  if FImage.Empty then
    Result := 0
  else
    Result := Image.Width;
end;


procedure TacHintImage.SetClientMargins(const Value: TacBordersSizes);
begin
  FClientMargins.Assign(Value);
end;


procedure TacHintImage.SetImage(const Value: TPngGraphic);
begin
  FImage.Assign(Value);
  ImageChanged;
  FOwner.FOwner.FOwner.Changed;
end;


procedure TacHintImage.SetOffsetHorz(const Value: integer);
begin
  FOffsetHorz := Value;
  FOwner.FOwner.FOwner.Changed;
end;


procedure TacHintImage.SetOffsetVert(const Value: integer);
begin
  FOffsetVert := Value;
  FOwner.FOwner.FOwner.Changed;
end;


procedure TacHintImage.SetImgHeight(const Value: integer);
begin
//
end;


procedure TacHintImage.SetImgWidth(const Value: integer);
begin
//
end;


procedure TacHintImage.SetBordersWidths(const Value: TacBordersSizes);
begin
  FBordersWidths.Assign(Value);
end;


procedure TacHintImage.ImageChanged;
var
  h, w: integer;
begin
  w := FImage.Width - 1;
  h := FImage.Height - 1;

  BordersWidths.FLeft  := w div 2;
  BordersWidths.FRight := BordersWidths.FLeft;
  ClientMargins.FLeft  := BordersWidths.FLeft;
  ClientMargins.FRight := BordersWidths.FLeft;

  BordersWidths.FTop    := h div 2;
  BordersWidths.FBottom := BordersWidths.FTop;
  ClientMargins.FTop    := BordersWidths.FTop;
  ClientMargins.FBottom := BordersWidths.FTop;
end;


procedure TacHintImage.AssignTo(Dest: TPersistent);
begin
  if Dest <> nil then begin
    TacHintImage(Dest).Image.Assign(Image);
    if FImageData.Size > 0 then begin
      FImageData.Seek(0, 0);
      TacHintImage(Dest).FImageData.LoadFromStream(FImageData);
    end
    else
      TacHintImage(Dest).FImageData.Size := 0;

    TacHintImage(Dest).OffsetHorz := OffsetHorz;
    TacHintImage(Dest).OffsetVert := OffsetVert;
    TacHintImage(Dest).ClientMargins  .Assign(ClientMargins);
    TacHintImage(Dest).BorderDrawModes.Assign(BorderDrawModes);
    TacHintImage(Dest).BordersWidths  .Assign(BordersWidths);
    TacHintImage(Dest).ShadowSizes    .Assign(ShadowSizes);
  end
  else
    inherited;
end;


procedure TacBorderDrawModes.AssignTo(Dest: TPersistent);
begin
  if Dest <> nil then begin
    TacBorderDrawModes(Dest).Top    := Top;
    TacBorderDrawModes(Dest).Left   := Left;
    TacBorderDrawModes(Dest).Bottom := Bottom;
    TacBorderDrawModes(Dest).Right  := Right;
    TacBorderDrawModes(Dest).Center := Center;
  end
  else
    inherited;
end;


constructor TacBorderDrawModes.Create(AOwner: TacHintImage);
begin
  FBottom := dmStretch;
  FLeft   := dmStretch;
  FTop    := dmStretch;
  FRight  := dmStretch;
  FCenter := dmStretch;
  FOwner  := AOwner;
end;


procedure TacBorderDrawModes.LoadFromStream(AStream: TStream);
begin
  AStream.Read(FTop,    SizeOf(FTop));
  AStream.Read(FLeft,   SizeOf(FLeft));
  AStream.Read(FRight,  SizeOf(FRight));
  AStream.Read(FBottom, SizeOf(FBottom));
  AStream.Read(FCenter, SizeOf(FCenter));
end;


procedure TacBorderDrawModes.SaveToStream(AStream: TStream);
begin
  AStream.Write(FTop,    SizeOf(FTop));
  AStream.Write(FLeft,   SizeOf(FLeft));
  AStream.Write(FRight,  SizeOf(FRight));
  AStream.Write(FBottom, SizeOf(FBottom));
  AStream.Write(FCenter, SizeOf(FCenter));
end;


procedure TacBorderDrawModes.SetDrawMode(const Index: Integer; const Value: TacBorderDrawMode);
begin
  case Index of
    0: FTop    := Value;
    1: FLeft   := Value;
    2: FBottom := Value;
    3: FRight  := Value;
    4: FCenter := Value;
  end;
  if FOwner.FOwner.FOwner <> nil then
    FOwner.FOwner.FOwner.FOwner.Changed;
end;


var
  HintStream: TResourceStream;


function TacCustomHintWindow.FullLayer: boolean;
begin
  if SkinIndex >= 0 then
    with DefaultManager.CommonSkinData.gd[SkinIndex] do
      Result := acLayered and ((Props[0].Transparency = 100) and (BorderIndex >= 0) or (OuterMode > 2) {Custom shadow})
  else
    Result := True;
end;


procedure TsAlphaHints.OnHideTimer(Sender: TObject);
begin
  HideHint;
  HideTimer.Enabled := False;
end;


procedure TsAlphaHints.OnCtrlTimer(Sender: TObject);
var
  Ctrl: TControl;
begin
  Ctrl := CtrlUnderMouse;
  if (Ctrl <> HintTimeCtrl) or (HintTimeCtrl = nil) then begin
    TacThreadedTimer(Sender).Enabled := False;
    HideTimer.Enabled := False;
    HideAll;
    HintTimeCtrl := nil;
  end;
end;


procedure TacHintImage.SetShadowSizes(const Value: TacBordersSizes);
begin
  FShadowSizes.Assign(Value);
end;


const
  ahAbbr: array [1..4] of AnsiChar = 'AShh';
  BufSize = 64;


function TsAlphaHints.LoadFromStream(AStream: TMemoryStream): TacHintTemplate;
var
  c: AnsiChar;
  i, n: integer;
  s: AnsiString;
  AStyle: TFontStyles;
  Buf: array [1..4] of AnsiChar;
  FontBuf: array [1..BufSize - 1] of AnsiChar;
begin
  AStream.Seek(0, 0);
  AStream.Read(Buf, SizeOf(Buf));
  if Buf = ahAbbr then begin
    Result := TacHintTemplate(Templates.Add);
    AStream.Read(c, 1);
    // Font
    AStream.Read(i, SizeOf(i));
    Result.Font.Color := TColor(i);
    if integer(Result.Font.Color) < 0 then
      Result.Font.Color := 0;

    AStream.Read(i, SizeOf(i));
    Result.Font.Size := i;
    AStream.Read(AStyle, SizeOf(AStyle));
    Result.Font.Style := AStyle;

    AStream.Read(n, SizeOf(n));
    AStream.Read(FontBuf, n);
    SetString(s, PAnsiChar(@FontBuf[1]), n);
    Result.Font.Name := s;

    Result.FImageDefault.LoadFromStream(AStream);
    if (c = CharOne) and (AStream.Position < AStream.Size) then begin // Other three images
      Result.FImageRightTop   .LoadFromStream(AStream);
      Result.FImageRightBottom.LoadFromStream(AStream);
      Result.FImageLeftBottom .LoadFromStream(AStream);
    end;
  end
  else begin
    Result := nil;
    ShowError('Unknown file format');
  end;
end;


procedure TsAlphaHints.SaveToStream(AStream: TMemoryStream);
var
  i, l: integer;
  s: AnsiString;
  AStyle: TFontStyles;
begin
  if (FTemplateIndex >= 0) and (FTemplateIndex < Templates.Count) then
    Template := Templates[FTemplateIndex]
  else
    Template := DefaultTemplate;

  if Template <> nil then begin
    AStream.Size := 0;
    AStream.Write(ahAbbr, SizeOf(ahAbbr));
    s := iff(Template.FImageRightTop <> nil, CharOne, ZeroChar);
    AStream.Write(AnsiChar(s[1]), 1);
    // Font
    if integer(Template.Font.Color) < 0 then begin
      i := 0;
      AStream.Write(i, SizeOf(Template.Font.Color))
    end
    else
      AStream.Write(Template.Font.Color, SizeOf(Template.Font.Color));

    i := Template.Font.Size;
    AStream.Write(i, SizeOf(i));

    AStyle := Template.Font.Style;
    AStream.Write(AStyle, SizeOf(AStyle));

    l := Length(Template.Font.Name);
    AStream.Write(l, SizeOf(l));
    s := Template.Font.Name;
    for i := 1 to l do
      AStream.Write(AnsiChar(s[i]), 1);

    // Images
    Template.FImageDefault.SaveToStream(AStream);
    if (Template.FImageRightTop <> nil) and not Template.SingleImageUsed then begin // Other three images
      Template.FImageRightTop   .SaveToStream(AStream);
      Template.FImageRightBottom.SaveToStream(AStream);
      Template.FImageLeftBottom .SaveToStream(AStream);
    end;
  end;
end;


procedure TacHintImage.SavetoStream(AStream: TStream);
var
  i: integer;
  c: AnsiChar;
  TmpStream: TMemoryStream;
begin
  BorderDrawModes.SaveToStream(AStream);
  BordersWidths  .SaveToStream(AStream);
  ClientMargins  .SaveToStream(AStream);
  ShadowSizes    .SaveToStream(AStream);
  if FImageData.Size = 0 then begin
    TmpStream := TMemoryStream.Create;
    try
      c := ZeroChar;
      AStream.Write(c, 1);
      FImage.SaveToStream(TmpStream);
      i := TmpStream.Size;
      AStream.Write(i, SizeOf(i));
      TmpStream.Seek(0, 0);
      TmpStream.SaveToStream(AStream);
    finally
      TmpStream.Free;
    end;
  end
  else begin
    c := CharOne;
    AStream.Write(c, 1);
    i := FImageData.Size;
    AStream.Write(i, SizeOf(i));
    FImageData.SaveToStream(AStream);
  end;
end;


procedure TsAlphaHints.SaveToFile(const FileName: string);
var
  OutStream: TMemoryStream;
begin
  if FileExists(FileName) then
    DeleteFile(FileName);

  OutStream := TMemoryStream.Create;
  try
    SaveToStream(OutStream);
    OutStream.SaveToFile(FileName);
  finally
    OutStream.Free;
  end;
end;


function TsAlphaHints.LoadFromFile(const FileName: string): TacHintTemplate;
var
  InStream: TMemoryStream;
  p: integer;
  s: string;
begin
  InStream := TMemoryStream.Create;
  try
    InStream.LoadFromFile(FileName);
    Result := LoadFromStream(InStream);
    if Result <> nil then begin
      s := ExtractFileName(FileName);
      p := Pos(s_Dot, s);
      Delete(s, p, Length(s) - p + 1);
      Result.Name := s;
    end;
  finally
    InStream.Free;
  end;
end;


procedure TacHintImage.LoadFromStream(AStream: TStream);
var
  l: integer;
  c: AnsiChar;
  Bmp: TBitmap;
  TmpStream: TMemoryStream;
begin
  BorderDrawModes.LoadFromStream(AStream);
  BordersWidths  .LoadFromStream(AStream);
  ClientMargins  .LoadFromStream(AStream);
  ShadowSizes    .LoadFromStream(AStream);
  AStream.Read(c, 1);
  AStream.Read(l, SizeOf(l));
  if FImage = nil then
    FImage := TPngGraphic.Create;

  if c = ZeroChar then begin // If bitmap
    TmpStream := TMemoryStream.Create;
    Bmp := TBitmap.Create;
    try
      TmpStream.CopyFrom(AStream, l);
      TmpStream.Seek(0, 0);
      Bmp.LoadFromStream(TmpStream);
      FImage.Assign(Bmp);
    finally
      FImageData.Size := 0;
      Bmp.Free;
      TmpStream.Free;
    end;
  end
  else begin
    FImageData.CopyFrom(AStream, l);
    FImageData.Seek(0, 0);
    FImage.LoadFromStream(FImageData);
  end;
end;


function TacHintImage.BitmapStored: boolean;
begin
  Result := FImageData.Size = 0;
end;


procedure TacHintImage.StreamChanged;
var
  aPng: TPngGraphic;
begin
  if FImageData.Size <> 0 then begin
    aPng := TPngGraphic.Create;
    try
      FImageData.Seek(0, 0);
      aPng.LoadFromStream(FImageData);
      FImage.Assign(aPng);
      FOwner.FOwner.FOwner.Changed;
    finally
      aPng.FreeImage;
      aPng.Free;
    end;
  end;
end;


procedure TacHintImage.DefineProperties(Filer: TFiler);
begin
  inherited DefineProperties(Filer);
  Filer.DefineBinaryProperty('ImgData', ReadData, WriteData, True);
end;


procedure TacHintImage.ReadData(Reader: TStream);
begin
  FImageData.LoadFromStream(Reader);
end;


procedure TacHintImage.WriteData(Writer: TStream);
begin
  FImageData.SaveToStream(Writer);
end;


function TacCustomHintWindow.CreateHintWindow(R: TRect): TForm;
begin
  Result := TForm.Create(nil);
  Result.Visible := False;
  Result.BorderStyle := bsNone;
  Result.Tag := ExceptTag;
  if SetWindowLong(Result.Handle, GWL_EXSTYLE, GetWindowLong(Result.Handle, GWL_EXSTYLE) or WS_EX_NOACTIVATE or WS_EX_LAYERED or WS_EX_TRANSPARENT or WS_EX_TOPMOST) = 0 then
    CheckLastError;

  if not SetWindowPos(Result.Handle, HWND_TOPMOST, R.Left, R.Top, WidthOf(R), HeightOf(R), SWP_NOACTIVATE or SWP_NOOWNERZORDER) then
    CheckLastError;

  if FullLayer then
    SetClassLong(Result.Handle, GCL_STYLE, GetClassLong(Result.Handle, GCL_STYLE) and not NCS_DROPSHADOW)
  else
    SetClassLong(Result.Handle, GCL_STYLE, GetClassLong(Result.Handle, GCL_STYLE) or NCS_DROPSHADOW);

  ShowWindow(Result.Handle, SW_SHOWNOACTIVATE);
end;


procedure TsAlphaHints.StartCheckTimer;
begin
  if CheckTimer <> nil then
    CheckTimer.Free;

  CheckTimer := TacThreadedTimer.Create(nil);
  CheckTimer.Enabled := False;
  CheckTimer.OnTimer := OnCheckTimer;
  CheckTimer.Interval := Application.HintShortPause + 300;
  CheckTimer.Enabled := True;
end;


procedure TsAlphaHints.StartHideTimer(HideTime: Cardinal = 0);
begin
  if HideTimer = nil then
    HideTimer := TacThreadedTimer.Create(Self);

  HideTimer.Enabled := False;
  if HideTime <> 0 then begin
    HideTimer.Interval := HideTime;
    HideTimer.OnTimer := OnHideTimer;
  end
  else begin
    HideTimer.Interval := 300;
    HideTimer.OnTimer := OnCtrlTimer;
  end;
  HideTimer.Enabled := True;
end;


procedure TsAlphaHints.OnCheckTimer(Sender: TObject);
begin
  if (Manager <> nil) and ((acHintWindow = nil) or not IsWindowVisible(acHintWindow.Handle) or not Manager.HintShowing) then begin // ????
    Blend := 0;
    FreeAndNil(AnimWindow);
    FreeAndNil(NewWindow);
    FreeAndNil(OldAlphaBmp);
    if Manager.FOldHintPause <> 0 then
      Application.HintPause := Manager.FOldHintPause;
  end;
  if CheckTimer <> nil then
    CheckTimer.Enabled := False;
end;


procedure TsShowTimer.Init(HintForm: TacCustomHintWindow; w, h: integer);
var
  FBlend: TBlendFunction;
begin
  Interval := acTimerInterval + AddInterval;
  OnTimer := IntOnTimer;
  Step := 0;
  StepCount := max({10 * }DefAnimationTime div acTimerInterval, 1) * StepDivider;
  CurrentHintForm := HintForm;
  InitBlendData(FBlend, 0);
  if Manager.Skinned and (DefaultManager.CommonSkinData.gd[CurrentHintForm.SkinIndex].Props[0].Transparency < 100) then
    Transparency := MulDiv(DefaultManager.CommonSkinData.gd[CurrentHintForm.SkinIndex].Props[0].Transparency, MaxByte, 100)
  else
    Transparency := 0;

  FBmpSize := MkSize(w, h);
  FBmpTopLeft := MkPoint;
  CurrentHintForm.CreateAlphaBmp(FBmpSize.cx, FBmpSize.cy);

  Transparency := Max(0, Transparency);
  StepB := (MaxByte - Transparency) / StepCount;
  SrcRect := Manager.AnimWindow.BoundsRect;

  if (CurrentHintForm <> nil) and (CurrentHintForm.FMousePos in [mpRightTop, mpRightBottom]) then
    dx := -(Manager.NewBounds.Left - SrcRect.Left) / Speed
  else
    dx := (Manager.NewBounds.Left - SrcRect.Left) / Speed;

  if (CurrentHintForm <> nil) and (CurrentHintForm.FMousePos in [mpLeftBottom, mpRightBottom]) then
    dy := -(Manager.NewBounds.Top - SrcRect.Top) / Speed
  else
    dy := (Manager.NewBounds.Top - SrcRect.Top) / Speed;

  x := SrcRect.Left;
  y := SrcRect.Top; 
  Enabled := True;
  if Manager.FOldHintPause = 0 then
    Manager.FOldHintPause := Application.HintPause;

  Application.HintPause := 100;
end;


procedure TsShowTimer.IntOnTimer(Sender: TObject);
begin
  if not (csDestroying in ComponentState) and
       (Manager.AnimWindow <> nil) and
         Manager.AnimWindow.HandleAllocated and
           (CurrentHintForm <> nil) and
             (CurrentHintForm.AlphaBmp <> nil) then
    with Manager do begin
      if (Step < StepCount) and (Blend <= MaxByte - Transparency - StepB) then begin
        inc(Step);
        Manager.Blend := Blend + StepB;
        if (dx <> 0) or (dy <> 0) then begin // Change position
          x := x + dx;
          y := y + dy;
          dx := (Manager.NewBounds.Left - x) / Speed;
          dy := (Manager.NewBounds.Top - y) / Speed;
          if NewWindow <> nil then // Recreate window if shadow mode must be changed
            if (GetClassLong(NewWindow.Handle, GCL_STYLE) and NCS_DROPSHADOW = NCS_DROPSHADOW) = CurrentHintForm.FullLayer then
              FreeAndNil(NewWindow);

          if NewWindow = nil then
            NewWindow := CurrentHintForm.CreateHintWindow(SrcRect);

          if OldAlphaBmp = nil then
            OldAlphaBmp := CreateBmp32;

          SetWindowPos(AnimWindow.Handle, HWND_TOPMOST, Round(x), Round(y), OldAlphaBmp.Width, OldAlphaBmp.Height, SWP_NOACTIVATE or SWP_NOOWNERZORDER);
          SetFormBlendValue(AnimWindow.Handle, OldAlphaBmp, MaxByte - Round(Blend));

          SetWindowPos(NewWindow.Handle, AnimWindow.Handle, Round(x), Round(y), CurrentHintForm.AlphaBmp.Width, CurrentHintForm.AlphaBmp.Height, SWP_NOACTIVATE or SWP_NOOWNERZORDER);
          SetFormBlendValue(NewWindow.Handle, CurrentHintForm.AlphaBmp, Round(Blend));
        end
        else begin // Change transparency only
          SetFormBlendValue(AnimWindow.Handle, CurrentHintForm.AlphaBmp, Round(Blend));
          ShowWindow(AnimWindow.Handle, SW_SHOWNOACTIVATE);
        end;
      end
      else begin // Finish
        Enabled := False;
        if NewWindow <> nil then begin // If was moved
          FreeAndNil(AnimWindow);
          AnimWindow := NewWindow;
          NewWindow := nil;
        end;
        SetFormBlendValue(AnimWindow.Handle, CurrentHintForm.AlphaBmp, MaxByte - Transparency);
        SetWindowPos(AnimWindow.Handle, HWND_TOPMOST, NewBounds.Left, NewBounds.Top, WidthOf(NewBounds), HeightOf(NewBounds), SWP_NOACTIVATE or SWP_NOOWNERZORDER);
        Blend := 0;
        if CurrentHintForm.AlphaBmp <> nil then begin // Save previous image for future animation
          if OldAlphaBmp <> nil then
            OldAlphaBmp.Free;

          OldAlphaBmp := CurrentHintForm.AlphaBmp;
          CurrentHintForm.AlphaBmp := nil;
        end;
      end;
  end;
end;


procedure TacCustomHintWindow.OnCloseHint;
begin
  if Manager <> nil then begin
    if Manager.ShowTimer <> nil then begin
      Manager.ShowTimer.Enabled := False;
      Manager.ResetHintInfo;
    end;
    if acHintWindow = Self then
      acHintWindow := nil;

    if Assigned(Manager.OnHideHint) then
      Manager.OnHideHint(Manager);

    Manager.StartCheckTimer;
  end;
end;


procedure TacPngHintWindow.DoDestroy;
begin
  if AlphaBmp <> nil then
    FreeAndNil(AlphaBmp);

  if (Manager <> nil) then
    FreeAndNil(Manager.AnimWindow);

  inherited;
end;


procedure TacPngHintWindow.UpdateImageTemplate;
begin
  case FMousePos of
    mpRightTop: begin
      if Template.FImageRightTop.Image.Empty then
        Template.FImageRightTop.StreamChanged;

      if not Template.FImageRightTop.Image.Empty then begin
        Image := Template.FImageRightTop;
        Manager.CheckAutoHintPos;
      end;
    end;

    mpLeftBottom: begin
      if Template.FImageLeftBottom.Image.Empty then
        Template.FImageLeftBottom.StreamChanged;

      if not Template.FImageLeftBottom.Image.Empty then begin
        Image := Template.FImageLeftBottom;
        Manager.CheckAutoHintPos;
      end;
    end;

    mpRightBottom: begin
      if Template.FImageRightBottom.Image.Empty then
        Template.FImageRightBottom.StreamChanged;

      if not Template.FImageRightBottom.Image.Empty then begin
        Image := Template.FImageRightBottom;
        Manager.CheckAutoHintPos;
      end;
    end;
  end;
end;


initialization
  DefaultTemplate := TacHintTemplate.Create(nil);
  HintStream := TResourceStream.Create(hInstance, 'ACHINT', RT_RCDATA);
  with DefaultTemplate.ImageDefault do begin
    DefaultTemplate.ImageDefault.Image.LoadFromStream(HintStream);
    BordersWidths.SetValues(11, 10, 14, 15);
    ShadowSizes  .SetValues( 6,  5,  7,  8);
    ClientMargins.SetValues(11,  8, 16, 15);
  end;
  FreeAndNil(HintStream);
  

finalization
  FreeAndNil(DefaultTemplate);

end.



