unit sDialogs;
{$I sDefs.inc}

interface

uses
  Windows, Dialogs, Forms, Classes, SysUtils, Graphics, ExtDlgs, Controls,
  {$IFDEF DELPHI_XE2} UITypes, {$ENDIF}
  {$IFDEF TNTUNICODE} TntDialogs, TntExtDlgs, TntFileCtrl, MultiMon, {$ENDIF}
  acntTypes, sConst;


type
{$IFNDEF NOTFORHELP}
  TsZipShowing = (zsAsFolder, zsAsFile);
{$ENDIF} // NOTFORHELP


{$IFDEF DELPHI_XE3}[ComponentPlatformsAttribute(pidWin32 or pidWin64)]{$ENDIF}
{$IFDEF TNTUNICODE}
  TsOpenDialog = class(TTntOpenDialog)
{$ELSE}
  TsOpenDialog = class(TOpenDialog)
{$ENDIF}
  private
    FZipShowing: TsZipShowing;
{$IFNDEF NOTFORHELP}
  public
    constructor Create(AOwner: TComponent); override;
{$ENDIF} // NOTFORHELP
  published
    property ZipShowing: TsZipShowing read FZipShowing write FZipShowing default zsAsFolder;
  end;


{$IFDEF DELPHI_XE3}[ComponentPlatformsAttribute(pidWin32 or pidWin64)]{$ENDIF}
{$IFDEF TNTUNICODE}
  TsOpenPictureDialog = class(TTntOpenPictureDialog)
{$ELSE}
  TsOpenPictureDialog = class(TOpenPictureDialog)
{$ENDIF}
{$IFNDEF NOTFORHELP}
  private
    FPicture: TPicture;
    function IsFilterStored: Boolean;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure DoSelectionChange; override;
    procedure DoShow; override;
  published
    property Filter stored IsFilterStored;
{$ENDIF} // NOTFORHELP
  end;


{$IFDEF DELPHI_XE3}[ComponentPlatformsAttribute(pidWin32 or pidWin64)]{$ENDIF}
{$IFDEF TNTUNICODE}
  TsSaveDialog = class(TTntSaveDialog)
{$ELSE}
  TsSaveDialog = class(TSaveDialog)
{$ENDIF}
{$IFNDEF NOTFORHELP}
  public
{$ENDIF} // NOTFORHELP
  end;


{$IFDEF DELPHI_XE3}[ComponentPlatformsAttribute(pidWin32 or pidWin64)]{$ENDIF}
{$IFDEF TNTUNICODE}
  TsSavePictureDialog = class(TTntSavePictureDialog)
{$ELSE}
  TsSavePictureDialog = class(TSavePictureDialog)
{$ENDIF}
{$IFNDEF NOTFORHELP}
  private
    FPicture: TPicture;
    function IsFilterStored: Boolean;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    property Filter stored IsFilterStored;
{$ENDIF} // NOTFORHELP
  end;


{$IFDEF DELPHI_XE3}[ComponentPlatformsAttribute(pidWin32 or pidWin64)]{$ENDIF}
  TsColorDialog = class(TColorDialog)
{$IFNDEF NOTFORHELP}
  private
    FUseAlpha,
    FStandardDlg: boolean;
    FMainColors: TStrings;
    FOnColorPreview: TacColorPreview;
    procedure SetMainColors(const Value: TStrings);
    procedure SetUseAlpha(const Value: boolean);
  public
    PreviewCmp: TComponent;
    function Execute: Boolean; override;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure DoShow; override;
    procedure DoClose; override;
  published
    property Options default [cdFullOpen];
{$ENDIF} // NOTFORHELP
    property MainColors: TStrings read FMainColors write SetMainColors;
    property UseAlpha: boolean read FUseAlpha write SetUseAlpha default False;
    property StandardDlg: boolean read FStandardDlg write FStandardDlg default False;
    property OnColorPreview: TacColorPreview read FOnColorPreview write FOnColorPreview;
  end;


{$IFNDEF NOTFORHELP}
function sCreateMessageDialog(const Title, Msg: acString; DlgType: TMsgDlgType; Buttons: TMsgDlgButtons): TForm;
  {$IFDEF D2010}
overload;
function sCreateMessageDialog(const Msg: acString; DlgType: TMsgDlgType; Buttons: TMsgDlgButtons; DefaultButton: TMsgDlgBtn): TForm; overload;
  {$ENDIF}
{$ENDIF} // NOTFORHELP

function sMessageDlg(const        Msg: acString; DlgType: TMsgDlgType; Buttons: TMsgDlgButtons; HelpCtx: LongInt): Integer; overload;
function sMessageDlg(const Title, Msg: acString; DlgType: TMsgDlgType; Buttons: TMsgDlgButtons; HelpCtx: LongInt): Integer; overload;
function sMessageDlgPos(const        Msg: acString; DlgType: TMsgDlgType; Buttons: TMsgDlgButtons; HelpCtx: LongInt; X, Y: Integer): Integer; overload;
function sMessageDlgPos(const Title, Msg: acString; DlgType: TMsgDlgType; Buttons: TMsgDlgButtons; HelpCtx: LongInt; X, Y: Integer): Integer; overload;

function sMessageDlgPosHelp(const Title, Msg: acString; DlgType: TMsgDlgType; Buttons: TMsgDlgButtons;
                            HelpCtx: LongInt; X, Y: Integer; const HelpFileName: acString): Integer;

procedure sShowMessage(const        Msg: acString); overload;
procedure sShowMessage(const Title, Msg: acString); overload;
procedure sShowMessageFmt(const        Msg: acString; const Params: array of const); overload;
procedure sShowMessageFmt(const Title, Msg: acString; const Params: array of const); overload;
procedure sShowMessagePos(const        Msg: acString; X, Y: Integer); overload;
procedure sShowMessagePos(const Title, Msg: acString; X, Y: Integer); overload;

{ Input dialog }
function sInputBox  (const ACaption, APrompt, ADefault: acString): acString;
function sInputQuery(const ACaption, APrompt: acString; var Value: acString): Boolean;

{$IFDEF TNTUNICODE}
function Application_MessageBoxW(const Text, Caption: PWideChar; Flags: Longint): Integer;
{$ENDIF}

{$IFNDEF NOTFORHELP}
function ShowCheckForm(const Title: string; const Text, CheckText: string; var aChecked: boolean): TModalResult;
{$ENDIF} // NOTFORHELP

implementation

uses
  sColorDialog, sSkinManager, acDials, sGraphUtils, acntUtils, sVCLUtils, sCheckBox, sBitBtn, acAlphaImageList, sLabel, acImage;


var
  Captions: array[TMsgDlgType] of acString = ('Warning', 'Error', 'Information', 'Confirm', '');


function sCreateMessageDialog(const Title, Msg: acString; DlgType: TMsgDlgType; Buttons: TMsgDlgButtons): TForm;
begin
{$IFDEF TNTUNICODE}
  Result := WideCreateMessageDialog(Msg, DlgType, Buttons);
{$else}
  Result := CreateMessageDialog(Msg, DlgType, Buttons);
{$ENDIF}
end;


{$IFDEF D2010}
function sCreateMessageDialog(const Msg: acString; DlgType: TMsgDlgType; Buttons: TMsgDlgButtons; DefaultButton: TMsgDlgBtn): TForm; overload;
begin
{$IFDEF TNTUNICODE}
  Result := WideCreateMessageDialog(Msg, DlgType, Buttons, DefaultButton);
{$else}
  Result := CreateMessageDialog(Msg, DlgType, Buttons, DefaultButton);
{$ENDIF}
end;
{$ENDIF}


function sMessageDlg(const Msg: acString; DlgType: TMsgDlgType; Buttons: TMsgDlgButtons; HelpCtx: LongInt): Integer;
begin
  Result := sMessageDlgPosHelp('', Msg, DlgType, Buttons, HelpCtx, -1, -1, '');
end;


function sMessageDlg(const Title, Msg: acString; DlgType: TMsgDlgType; Buttons: TMsgDlgButtons; HelpCtx: LongInt): Integer;
begin
  Result := sMessageDlgPosHelp(Title, Msg, DlgType, Buttons, HelpCtx, -1, -1, '');
end;


function sMessageDlgPos(const Msg: acString; DlgType: TMsgDlgType; Buttons: TMsgDlgButtons; HelpCtx: LongInt; X, Y: Integer): Integer;
begin
  Result := sMessageDlgPosHelp('', Msg, DlgType, Buttons, HelpCtx, X, Y, '');
end;


function sMessageDlgPos(const Title, Msg: acString; DlgType: TMsgDlgType; Buttons: TMsgDlgButtons; HelpCtx: LongInt; X, Y: Integer): Integer;
begin
  Result := sMessageDlgPosHelp(Title, Msg, DlgType, Buttons, HelpCtx, X, Y, '');
end;


{$IFDEF TNTUNICODE}
function Application_MessageBoxW(const Text, Caption: PWideChar; Flags: Longint): Integer;
var
  ActiveWindow, TaskActiveWindow: HWnd;
  MBMonitor, AppMonitor: HMonitor;
  FocusState: TFocusState;
  MonInfo: TMonitorInfo;
  WindowList: Pointer;
  Rect: TRect;
begin
  with Application do begin
  {$IFDEF D2006}
    ActiveWindow := ActiveFormHandle;
  {$ELSE}
    ActiveWindow := GetActiveWindow;
  {$ENDIF}
    if ActiveWindow = 0 then
      TaskActiveWindow := Handle
    else
      TaskActiveWindow := ActiveWindow;

    MBMonitor := MonitorFromWindow(ActiveWindow, MONITOR_DEFAULTTONEAREST);
    AppMonitor := MonitorFromWindow(Handle, MONITOR_DEFAULTTONEAREST);
    if MBMonitor <> AppMonitor then begin
      MonInfo.cbSize := Sizeof(TMonitorInfo);
      GetMonitorInfo(MBMonitor, @MonInfo);
      GetWindowRect(Handle, Rect);
      with MonInfo.rcMonitor do
        SetWindowPos(Handle, 0, Left + (Right - Left) div 2, Top + (Bottom - Top) div 2, 0, 0, SWP_NOACTIVATE or SWP_NOREDRAW or SWP_NOSIZE or SWP_NOZORDER);
    end;
    WindowList := DisableTaskWindows(ActiveWindow);
    FocusState := SaveFocusState;
    if UseRightToLeftReading then
      Flags := Flags or MB_RTLREADING;

    try
{$IFDEF DELPHI7UP}
      Application.ModalStarted;
{$ENDIF}
      Result := MessageBoxW(TaskActiveWindow, Text, Caption, Flags);
{$IFDEF DELPHI7UP}
      Application.ModalFinished;
{$ENDIF}
    finally
      if MBMonitor <> AppMonitor then
        SetWindowPos(Handle, 0,
                     Rect.Left + (WidthOf(Rect)  div 2),
                     Rect.Top  + (HeightOf(Rect) div 2),
                     0, 0, SWP_NOACTIVATE or SWP_NOREDRAW or SWP_NOSIZE or SWP_NOZORDER);
      EnableTaskWindows(WindowList);
      SetActiveWindow(ActiveWindow);
      RestoreFocusState(FocusState);
    end;
  end;
end;
{$ENDIF}


procedure MsgBoxCallback(var lpHelpInfo: THelpInfo); stdcall;
begin
  if lpHelpInfo.dwContextId <> 0 then
    Application.HelpContext(lpHelpInfo.dwContextId);
end;


function sMessageDlgPosHelp(const Title, Msg: acString; DlgType: TMsgDlgType; Buttons: TMsgDlgButtons; HelpCtx: LongInt; X, Y: Integer; const HelpFileName: acString): Integer;
const
  MB_HELP = $4000;
  MB_YESTOALL = $A000;
var
  Flags: DWORD;
  Caption: acString;
  ActiveWnd: hwnd;
{$IFDEF TNTUNICODE}
  mParams: TMsgBoxParamsW;
{$ELSE}
  mParams: TMsgBoxParams;
{$ENDIF}
  WindowList: Pointer;
begin
  case DlgType of
    mtWarning:      Flags := MB_ICONWARNING;
    mtError:        Flags := MB_ICONERROR;
    mtInformation:  Flags := MB_ICONINFORMATION;
    mtConfirmation: Flags := MB_ICONQUESTION
    else            Flags := 0;
  end;
  Caption := iff(Title = '', Captions[DlgType], Title);

  if mbOk in Buttons then
    Flags := Flags or DWord(iff(mbCancel in Buttons, MB_OKCANCEL, MB_OK))
  else
    if (mbAbort in Buttons) or (mbIgnore in Buttons) then
      Flags := Flags or MB_ABORTRETRYIGNORE
    else
      if (mbYes in Buttons) or (mbNo in Buttons) then
        Flags := Flags or DWord(iff(mbCancel in Buttons, MB_YESNOCANCEL, MB_YESNO))
      else
        if mbRetry in Buttons then
          Flags := Flags or MB_RETRYCANCEL;

  if mbHelp in Buttons then
    Flags := Flags or MB_HELP;

  if Application.BiDiMode = bdRightToLeft then
    Flags := Flags {or MB_RTLREADING} or MB_RIGHT;

  DlgLeft := X;
  DlgTop := Y;
{$IFDEF DELPHI7UP}
  Application.ModalStarted;
{$ENDIF}
  FillChar(mParams, SizeOf(mParams), 0);
  mParams.cbSize := SizeOf(mParams);
  mParams.dwContextHelpId := acNativeUInt(HelpCtx);
  mParams.dwStyle := Flags;
  mParams.lpszCaption := PacChar(Caption);
  mParams.lpszText := PacChar(Msg);
{$IFDEF D2005}
  mParams.hwndOwner := Application.ActiveFormHandle;
{$ELSE}
  mParams.hwndOwner := Application.Handle;
{$ENDIF}
{$T-}
  mParams.lpfnMsgBoxCallback := @MsgBoxCallback;
{$T+}
  ActiveWnd := GetActiveWindow;
  WindowList := DisableTaskWindows(0);
{$IFDEF TNTUNICODE}
  Result := integer(MessageBoxIndirectW(mParams));
{$ELSE}
  Result := integer(MessageBoxIndirect(mParams));
{$ENDIF}
  EnableTaskWindows(WindowList);
  if ActiveWnd <> 0 then begin
    SetFocus(0);
    SetFocus(ActiveWnd);
  end;
{$IFDEF DELPHI7UP}
  Application.ModalFinished;
{$ENDIF}
  DlgLeft := -1;
  DlgTop := -1;
end;


procedure sShowMessage(const Msg: acString);
begin
  sShowMessagePos(Msg, -1, -1);
end;


procedure sShowMessage(const Title, Msg: acString);
begin
  sShowMessagePos(Title, Msg, -1, -1);
end;


procedure sShowMessageFmt(const Msg: acString; const Params: array of const);
begin
  sShowMessage(Format(Msg, Params));
end;


procedure sShowMessageFmt(const Title, Msg: acString; const Params: array of const);
begin
  sShowMessage(Title, Format(Msg, Params));
end;


procedure sShowMessagePos(const Msg: acString; X, Y: Integer);
begin
  sMessageDlgPos(Msg, mtCustom, [mbOK], 0, X, Y);
end;


procedure sShowMessagePos(const Title, Msg: acString; X, Y: Integer);
begin
  sMessageDlgPos(Title, Msg, mtCustom, [mbOK], 0, X, Y);
end;


function sInputQuery(const ACaption, APrompt: acString; var Value: acString): Boolean;
begin
{$IFDEF TNTUNICODE}
  Result := WideInputQuery(ACaption, APrompt, Value);
{$ELSE}
  Result := InputQuery(ACaption, APrompt, Value);
{$ENDIF}
end;


function sInputBox(const ACaption, APrompt, ADefault: acString): acString;
begin
  Result := ADefault;
  sInputQuery(ACaption, APrompt, Result);
end;


type
  TacWarningForm = class(TForm);

function ShowCheckForm(const Title: string; const Text, CheckText: string; var aChecked: boolean): TModalResult;
const
  Margin = 30;
var
  f: TacWarningForm;
  cb: TsCheckBox;
  btn: TsBitBtn;
{$IFNDEF ALITE}
  cl: TsCharImageList;
  img: TsImage;
  l: TsHTMLLabel;
{$ELSE}
  l: TsLabel;
{$ENDIF}
begin
  f := TacWarningForm.CreateNew(Application);
  try
    with f do begin
      BorderStyle := bsDialog;
      Position := poScreenCenter;
      Caption := Title;
      Width := 550;
      Height := 170;
    end;
{$IFNDEF ALITE}
    cl := TsCharImageList.Create(f); // F06A   F071
    cl.Width := 48;
    cl.Height := 48;
    with TacCharListItem(cl.Items.Add) do begin
      Color := TColor($FF0080FF);
      Char := $F071;
    end;
    img := TsImage.Create(f);
    with img do begin
      AutoSize := True;
      Images := cl;
      ImageIndex := 0;
      Left := Margin;
      Top := Margin;
      Parent := f;
    end;
    l := TsHTMLLabel.Create(f);
{$ELSE}
    l := TsLabel.Create(f);
{$ENDIF}
    with l do begin
      AutoSize := False;
      Left := {$IFNDEF ALITE}img.BoundsRect.Right +{$ENDIF} Margin;
      Width := f.Width - Left - Margin;
      Top := Margin;
      Height := 30;
      Parent := f;
      Caption := Text;
    end;
    cb := TsCheckBox.Create(f);
    with cb do begin
      AutoSize := False;
      Left := l.Left;
      Width := 280;
      Top := 60;
      Height := 50;
      Parent := f;
      Checked := aChecked;
      WordWrap := True;
      TextIndent := 6;
      Caption := CheckText;
    end;
    btn := TsBitBtn.Create(f);
    with btn do begin
      ModalResult := mrOk;
      Default := True;
      Top := cb.Top - (Height - cb.Height) div 2;
      Width := 80;
      Left := f.Width - 70 - btn.Width;
      Caption := 'Ok';
      Parent := f;
    end;
    Result := f.ShowModal;
    aChecked := cb.Checked;
  finally
    f.Free;
  end;
end;


constructor TsOpenDialog.Create(AOwner: TComponent);
begin
  inherited;
  FZipShowing := zsAsFolder;
end;


constructor TsOpenPictureDialog.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Filter := GraphicFilter(TGraphic);
end;


destructor TsOpenPictureDialog.Destroy;
begin
  if Assigned(FPicture) then
    FreeAndNil(FPicture);

  inherited;
end;


procedure TsOpenPictureDialog.DoSelectionChange;
begin
  if not (csDestroying in ComponentState) then
    inherited DoSelectionChange;
end;


procedure TsOpenPictureDialog.DoShow;
begin
  inherited DoShow;
end;


function TsOpenPictureDialog.IsFilterStored: Boolean;
begin
  Result := Filter <> GraphicFilter(TGraphic);
end;


constructor TsSavePictureDialog.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Filter := GraphicFilter(TGraphic);
end;


destructor TsSavePictureDialog.Destroy;
begin
  if Assigned(FPicture) then
    FreeAndNil(FPicture);

  inherited;
end;


function TsSavePictureDialog.IsFilterStored: Boolean;
begin
  Result := Filter <> GraphicFilter(TGraphic);
end;


constructor TsColorDialog.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FMainColors := TStringList.Create;
  FStandardDlg := False;
  FUseAlpha := False;
  Options := [cdFullOpen];
  PreviewCmp := nil;
end;


destructor TsColorDialog.Destroy;
begin
  FreeAndNil(FMainColors);
  inherited Destroy;
end;


procedure TsColorDialog.DoClose;
begin
  inherited;
end;


procedure TsColorDialog.DoShow;
begin
  inherited;
end;


function TsColorDialog.Execute: Boolean;
var
  i: integer;
  C: TColor;
begin
  if not FStandardDlg then begin
    sColorDialogForm := TsColorDialogForm.Create(Application);
    sColorDialogForm.OnColorPreview := OnColorPreview;
    sColorDialogForm.AddPal.UseAlpha := FUseAlpha;
    sColorDialogForm.MainPal.UseAlpha := FUseAlpha;
    if PreviewCmp = nil then
      sColorDialogForm.PreviewCmp := Self
    else
      sColorDialogForm.PreviewCmp := PreviewCmp;

    with sColorDialogForm do begin
      Owner := Self;
      if CustomColors.Count > 0 then begin
        for i := 0 to CustomColors.Count - 1 do
          if i < AddPal.Colors.Count then
            AddPal.Colors[i] := CustomColors[i];
      end
      else
        AddPal.Colors.Assign(acCustomColors);

      AddPal.GenerateColors;
      if MainColors.Count > 0 then begin
        MainPal.Colors.Assign(MainColors);
        MainPal.GenerateColors;
      end;
      ModalResult := mrCancel;
      BorderStyle := bsSingle;
      UseAlpha := Self.UseAlpha;
      InitControls(cdFullOpen in Options, cdShowHelp in Options, Self.Color, bsSingle);
      ShowModal;
      Result := ModalResult = mrOk;
      CustomColors.Assign(AddPal.Colors);
      DoClose;
      if Result then
        TsColorDialog(Self).Color := UpdateColorAlpha(TColor(SelectedPanel.Tag))
      else
        if Assigned(OnColorPreview) then begin
          C := TsColorDialog(Self).Color;
          OnColorPreview(Self, TsColorDialog(Self).Color, TColor(SelectedPanel.Tag), C);
        end;

      FreeAndNil(sColorDialogForm);
    end;
  end
  else
    Result := inherited Execute;
end;


procedure TsColorDialog.SetMainColors(const Value: TStrings);
begin
  FMainColors.Assign(Value);
end;


procedure TsColorDialog.SetUseAlpha(const Value: boolean);
begin
  FUseAlpha := Value;
  if Value then
    Color := TColor(Cardinal(Color) or $FF000000)
  else
    Color := Color and $FFFFFF;
end;

end.
