 unit sToolBar;
{$I sDefs.inc}
//{$DEFINE LOGGED}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, ToolWin, ComCtrls,
  {$IFNDEF DELPHI5} Types, {$ENDIF}
  {$IFDEF TNTUNICODE} TntComCtrls, {$ENDIF}
  sCommonData, sConst, sDefaults;


type
{$IFDEF DELPHI_XE3}[ComponentPlatformsAttribute(pidWin32 or pidWin64)]{$ENDIF}
{$IFDEF TNTUNICODE}
  TsToolBar = class(TTntToolBar)
{$ELSE}
  TsToolBar = class(TToolBar)
{$ENDIF}
{$IFNDEF NOTFORHELP}
  private
    HotButtonIndex: integer;
    FCommonData: TsCommonData;
    FDisabledKind: TsDisabledKind;
    procedure WMNCPaint(var Message: TWMNCPaint); message WM_NCPAINT;
    procedure SetDisabledKind(const Value: TsDisabledKind);
  protected
    DroppedButton: TToolButton;
    procedure WndProc(var Message: TMessage); override;
    procedure CancelMenu; override;
    function OffsetX: integer;
    function OffsetY: integer;
    function PrepareCache: boolean;
    function TextIsVisible(Button: TToolButton): boolean;
    procedure ChangeScale(M, D: Integer{$IFDEF DELPHI_10BERLIN}; isDpiChange: Boolean{$ENDIF}); override;
    procedure OurAdvancedCustomDraw(Sender: TToolBar; const ARect: TRect; Stage: TCustomDrawStage; var DefaultDraw: Boolean);
    procedure OurAdvancedCustomDrawButton(Sender: TToolBar; Button: TToolButton; State: TCustomDrawState;
                                          Stage: TCustomDrawStage; var Flags: TTBCustomDrawFlags; var DefaultDraw: Boolean);
    function GetButtonRect(Index: integer): TRect;
    function IndexByMouse(MousePos: TPoint): integer;
    procedure RepaintButton(Index: integer);
  {$IFDEF DELPHI6UP}
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  {$ELSE}
    procedure StdPaint(Button: TToolButton; State: TCustomDrawState; Bmp: TBitmap);
  {$ENDIF}
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure AfterConstruction; override;
    procedure Loaded; override;
    procedure UpdateEvents;
  published
    property Flat;
{$ENDIF} // NOTFORHELP
    property BtnDisabledKind: TsDisabledKind read FDisabledKind write SetDisabledKind default DefDisabledKind;
    property SkinData: TsCommonData read FCommonData write FCommonData;
  end;


implementation

uses
  CommCtrl, ImgList, Menus,
  {$IFDEF LOGGED} sDebugMsgs, {$ENDIF}
  acntUtils, sStyleSimply, sVCLUtils, sMessages, sMaskData, sGraphUtils,
  sAlphaGraph, sSkinManager, sThirdParty;


procedure PaintSeparator(const DC: hdc; const R: TRect; const ToolBar: TsToolBar; IsDivider: boolean);
var
  SavedDC: hdc;
  Bmp: TBitmap;
  si, mi: integer;
  nRect, cR: TRect;
  BGInfo: TacBGInfo;
begin
  si := ToolBar.SkinData.SkinManager.SkinCommonInfo.Sections[ssDivider];
  mi := ToolBar.SkinData.CommonSkinData.gd[si].BorderIndex;
  if ToolBar.SkinData.SkinManager.IsValidImgIndex(mi) then begin
    Bmp := CreateBmp32(R);
    BGInfo.BgType := btUnknown;
    BGInfo.Offset := MkPoint;
    BGInfo.Bmp := nil;
    BGInfo.R := MkRect(Bmp);
    BGInfo.PleaseDraw := False;
    GetBGInfo(@BGInfo, ToolBar.Handle, False);
    if BGInfo.BgType = btCache then
      BitBlt(Bmp.Canvas.Handle, 0, 0, Bmp.Width, Bmp.Height, BGInfo.Bmp.Canvas.Handle, BGInfo.Offset.X + R.Left, BGInfo.Offset.Y + R.Top, SRCCOPY)
    else
      FillDC(Bmp.Canvas.Handle, MkRect(Bmp), BGInfo.Color);

    nRect.Top    := 2;
    nRect.Left   := Bmp.Width div 2 - 2;
    nRect.Bottom := Bmp.Height - 2;
    nRect.Right  := nRect.Left + 4;

    if IsDivider then
      PaintItem(si, BGInfoToCI(@BGInfo), True, 0, nRect, Point(R.Left + ToolBar.BorderWidth, R.Top + ToolBar.BorderWidth), Bmp, ToolBar.SkinData.CommonSkinData);

    if DC = ToolBar.SkinData.PrintDC then begin
      SavedDC := SaveDC(DC);
      cR := ACClientRect(ToolBar.Handle);
      MoveWindowOrg(DC, -cR.Left, -cR.Top);
    end
    else
      SavedDC := 0;

    BitBlt(DC, ToolBar.OffsetX + R.Left, ToolBar.OffsetY + R.Top, Bmp.Width, Bmp.Height, Bmp.Canvas.Handle, 0, 0, SRCCOPY);
    if DC = ToolBar.SkinData.PrintDC then
      RestoreDC(DC, SavedDC);

    FreeAndNil(Bmp);
  end;
end;


function TsToolBar.OffsetX: integer;
begin
  Result := (integer(ebLeft in EdgeBorders) + BorderWidth) * integer(EdgeInner <> esNone) + (integer(ebLeft in EdgeBorders) + BorderWidth) * integer(EdgeOuter <> esNone)
end;


function TsToolBar.OffsetY: integer;
begin
  Result := (integer(ebTop in EdgeBorders) + BorderWidth) * integer(EdgeInner <> esNone) + (integer(ebTop in EdgeBorders) + BorderWidth) * integer(EdgeOuter <> esNone)
end;


procedure TsToolBar.AfterConstruction;
begin
  inherited;
  FCommonData.Loaded;
  UpdateEvents;
end;


{$IFDEF DELPHI6UP}

procedure TsToolBar.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if not (csDesigning in ComponentState) and Assigned(FCommonData) and FCommonData.Skinned then
    if (AComponent is TPopupMenu) and Assigned(DefaultManager) and (Operation = opInsert) then
      DefaultManager.SkinableMenus.HookPopupMenu(TPopupMenu(AComponent), True);
end;

{$ELSE}

procedure TsToolBar.StdPaint(Button: TToolButton; State: TCustomDrawState; Bmp: TBitmap);
const
  DownStyles: array[Boolean] of Integer = (BDR_RAISEDINNER, BDR_SUNKENOUTER);
  FillStyles: array[Boolean] of Integer = (BF_MIDDLE, 0);
var
  PaintRect: TRect;
  Mode, DrawFlags: Integer;
begin
  if not Button.Marked and not Button.Indeterminate and ((State = []) or (State = [cdsDisabled])) then
    Mode := integer(cdsHot in State)
  else
    Mode := 1 + integer(([cdsSelected, cdsChecked] * State <> []) or Button.Marked or Button.Indeterminate);

  PaintRect := Button.ClientRect;
  if not Flat then begin
    DrawFlags := DFCS_BUTTONPUSH or DFCS_ADJUSTRECT;
    if Mode = 2 then
      DrawFlags := DrawFlags or DFCS_PUSHED;

    DrawFrameControl(Bmp.Canvas.Handle, PaintRect, DFC_BUTTON, DrawFlags);
  end
  else begin
    if (Mode > 0) or (csDesigning in ComponentState) then
      DrawEdge(Bmp.Canvas.Handle, PaintRect, DownStyles[Mode = 2], FillStyles[Transparent] or BF_RECT)
    else
      if not Transparent then begin
        Bmp.Canvas.Brush.Color := Color;
        Bmp.Canvas.FillRect(PaintRect);
      end;

    InflateRect(PaintRect, -1, -1);
  end;
{$IFDEF FPS}
  if FState in [bsDown, bsExclusive] then
    if (FState = bsExclusive) and (not Flat or not MouseInControl) then begin
      Bmp.Canvas.Brush.Bitmap := AllocPatternBitmap(clBtnFace, clBtnHighlight);
      Bmp.Canvas.FillRect(PaintRect);
    end;
{$ENDIF}
end;

{$ENDIF}

procedure TsToolBar.CancelMenu;
var
  i: integer;
begin
  if DroppedButton <> nil then begin
    i := DroppedButton.Index;
    DroppedButton := nil;
    RepaintButton(i);
  end;
  inherited;
end;


procedure TsToolBar.ChangeScale(M, D: Integer{$IFDEF DELPHI_10BERLIN}; isDpiChange: Boolean{$ENDIF});
begin
  Font.Height := MulDiv(Font.Height, M, D);
  Height := MulDiv(Height, M, D);
{$IFDEF D2005}
  ButtonHeight := MulDiv(ButtonHeight, M, D);
  ButtonWidth := MulDiv(ButtonWidth, M, D);
{$ENDIF}
end;


constructor TsToolBar.Create(AOwner: TComponent);
begin
  FCommonData := TsCommonData.Create(Self, True);
  FCommonData.COC := COC_TsToolBar;
  inherited Create(AOwner);
  FDisabledKind := DefDisabledKind;
  DoubleBuffered := True;
  HotButtonIndex := -1;
end;


destructor TsToolBar.Destroy;
begin
  FreeAndNil(FCommonData);
  inherited Destroy;
end;


function TsToolBar.GetButtonRect(Index: integer): TRect;
begin
  if SendMessage(Handle, TB_GETITEMRECT, Index, LPARAM(@Result)) = 0 then
    Result.Left := Result.Right;
end;


function TsToolBar.IndexByMouse(MousePos: TPoint): integer;
var
  i: integer;
begin
  Result := -1;
  for i := 0 to ButtonCount - 1 do
    if PtInRect(GetButtonRect(i), MousePos) then begin
      if (TControl(Buttons[I]) is TToolButton) and (Buttons[i].Style in [tbsButton, tbsCheck, tbsDropDown]) then
        Result := i;

      Exit;
    end;
end;


procedure TsToolBar.Loaded;
begin
  inherited;
  FCommonData.Loaded;
  UpdateEvents;
end;


procedure CopyCache(Control: TWinControl; SkinData: TsCommonData; SrcRect, DstRect: TRect; DstDC: HDC);
var
  SaveIndex: HDC;
  i: integer;
  Designing: boolean;
begin
  SaveIndex := SaveDC(DstDC);
  IntersectClipRect(DstDC, DstRect.Left, DstRect.Top, DstRect.Right, DstRect.Bottom);
  Designing := csDesigning in Control.ComponentState;
  try
    for i := 0 to Control.ControlCount - 1 do
      with Control, Controls[i] do begin
        if (Controls[i] is TToolButton) and (csDesigning in ComponentState) then
          Continue;

        if (Controls[i] is TGraphicControl) and (SkinData.SkinManager <> nil) and SkinData.SkinManager.Options.StdImgTransparency then
          Continue;

        if not Visible or not RectIsVisible(DstRect, BoundsRect) then
          Continue;

        if (csOpaque in ControlStyle) or (Controls[i] is TGraphicControl) or Designing then
          ExcludeClipRect(DstDC, Left, Top, Left + Width, Top + Height);
      end;

    BitBlt(DstDC, DstRect.Left, DstRect.Top, WidthOf(DstRect), HeightOf(DstRect), SkinData.FCacheBmp.Canvas.Handle, SrcRect.Left, SrcRect.Top, SRCCOPY);
  finally
    RestoreDC(DstDC, SaveIndex);
  end;
end;


procedure TsToolBar.OurAdvancedCustomDraw(Sender: TToolBar; const ARect: TRect; Stage: TCustomDrawStage; var DefaultDraw: Boolean);
var
  RC, RW: TRect;
  DC: hdc;
begin
  if FCommonData.Skinned then begin
    DC := Canvas.Handle;
    FCommonData.Updating := FCommonData.Updating;
    if not (Stage in [cdPrePaint]) then
      DefaultDraw := False
    else
      if not FCommonData.BGChanged or PrepareCache then begin
        FCommonData.FCacheBMP.Canvas.Font.Assign(Font);
        Windows.GetClientRect(Handle, RC);
        GetWindowRect(Handle, RW);
        MapWindowPoints(0, Handle, RW, 2);
        OffsetRect(RC, -RW.Left, -RW.Top);

        CopyCache(Self, FCommonData, RC, ARect, DC);
        sVCLUtils.PaintControls(DC, Self, True, MkPoint);
        SetParentUpdated(Self);
        if (RC.Left > 0) or (RC.Top > 0) or (RC.Right <> Width) or (RC.Bottom <> Height) then
          Perform(WM_NCPAINT, 0, 0);
      end
      else
        DefaultDraw := False;
  end
  else begin
    DefaultDraw := True;
    inherited;
  end
end;


procedure TsToolBar.OurAdvancedCustomDrawButton(Sender: TToolBar; Button: TToolButton; State: TCustomDrawState; Stage: TCustomDrawStage; var Flags: TTBCustomDrawFlags; var DefaultDraw: Boolean);
var
  R, iR: TRect;
  ci: TCacheInfo;
  BtnBmp: TBitmap;
  bWidth, bHeight, Mode, iSkinIndex: integer;

  function AddedWidth: integer;
  begin
    if Button.Style = tbsDropDown then
      Result := GetScrollSize(Skindata.SkinManager) - 2
    else
      Result := 0
  end;

  function IntButtonWidth: integer;
  begin
    Result := Button.Width - AddedWidth;
  end;

  function ButtonWidth: integer;
  begin
    Result := Button.Width;
  end;

  function ImgRect: TRect;
  begin
    with Result do
      if not List then begin
        Left   := (IntButtonWidth - Images.Width) div 2 + 1;
        Top    := (Button.Height - Images.Height - integer(TextIsVisible(Button)) * (FCommonData.FCacheBMP.Canvas.TextHeight('A') + 3)) div 2;
        Right  := Result.Left + Images.Width;
        Bottom := Result.Bottom + Images.Height;
      end
      else begin
        Left   := 5;
        Top    := (Button.Height - Images.Height) div 2;
        Right  := Result.Left + Images.Width;
        Bottom := Result.Bottom + Images.Height;
      end;

    if Mode = 2 then
      if SkinData.Skinned and SkinData.SkinManager.ButtonsOptions.ShowFocusRect then
        OffsetRect(Result, 1, 1);
  end;

  procedure DrawBtnCaption;
  var
    cRect: TRect;
    Ndx: integer;
    pc: PacChar;

    function CaptionRect: TRect;
    var
      l, t, r, b, dh: integer;
    begin
      if not List then begin
        l := (IntButtonWidth - FCommonData.FCacheBMP.Canvas.TextWidth(Button.Caption)) div 2;
        if Assigned(Images) then begin
          dh := (Button.Height - Images.Height - FCommonData.FCacheBMP.Canvas.TextHeight('A') - 3) div 2;
          t := dh + Images.Height + 3;
        end
        else begin
          dh := (Button.Height - FCommonData.FCacheBMP.Canvas.TextHeight('A')) div 2;
          t := dh;
        end;
        r := IntButtonWidth - l;
        b := Button.Height - dh;
        Result := Rect(l - 1, t, r + 2, b);
      end
      else begin
        if Assigned(Images) and (Button.ImageIndex >= 0) then
          Result.Left := 6 + Images.Width
        else
          Result.Left := 1;

        Result.Right := IntButtonWidth - 2;
        Result.Top := 2;
        Result.Bottom := Button.Height - 2;
      end;
      if Mode = 2 then
        if SkinData.Skinned and SkinData.SkinManager.ButtonsOptions.ShowFocusRect then
          OffsetRect(Result, 1, 1);
    end;

  begin
    if TextIsVisible(Button) then begin
      cRect := CaptionRect;
{$IFDEF TNTUNICODE}
      if Button is TTntToolButton then
        pc := PacChar(TTntToolButton(Button).Caption)

      else
{$ENDIF}
{$WARNINGS OFF}
        pc := PacChar(Button.Caption);
{$WARNINGS ON}

      if NeedParentFont(SkinData.SkinManager, iSkinIndex, Mode) then
        Ndx := GetFontIndex(Self, SkinData.SkinIndex, SkinData.SkinManager)
      else
        Ndx := iSkinIndex;

      acWriteTextEx(BtnBMP.Canvas, pc, Enabled and (DisabledImages = nil), cRect, DT_CENTER or DT_VCENTER or DT_SINGLELINE, Ndx, Mode > 0)
    end;
  end;

  procedure DrawBtnGlyph;
  var
    i: integer;
  begin
    i := integer(Mode > 0);
    BtnBmp.Canvas.Font.Color := FCommonData.CommonSkinData.gd[iSkinIndex].Props[i].FontColor.Color;
    if Assigned(Images) and IsValidIndex(Button.ImageIndex, GetImageCount(Images)) then
      sThirdParty.CopyToolBtnGlyph(Self, Button, State, Stage, Flags, BtnBmp);
  end;

  procedure DrawArrow;
  var
    i, Mode, x, y: integer;
  begin
    with FCommonData.SkinManager, FCommonData.CommonSkinData, gd[iSkinIndex] do begin
      if (DroppedButton = Button) and Assigned(Button.DropDownMenu) or Button.Down then
        Mode := 2
      else
        Mode := integer(cdsHot in State);

      R.Left := IntButtonWidth;
      R.Right := Button.Width;
      if IsValidImgIndex(BorderIndex) then
        DrawSkinRect(BtnBmp, R, ci, ma[BorderIndex], Mode, True);

      with ScrollBtns[asBottom] do
        if IsValidIndex(GlyphIndex, High(ma)) then begin
          x := IntButtonWidth + (AddedWidth - 3 - ma[GlyphIndex].Width) div 2 + 2;
          y := (ButtonHeight - ma[GlyphIndex].Height) div 2;
          DrawSkinGlyph(BtnBmp, Point(x, y), Mode, 1, ma[GlyphIndex], MakeCacheInfo(BtnBmp));
        end
        else begin
          i := integer(Mode > 0);
          DrawColorArrow(BtnBmp, gd[iSkinIndex].Props[i].FontColor.Color, R, asBottom, GetPPI(SkinData));
        end;
    end;
  end;

begin
  if FCommonData.Skinned {$IFNDEF DELPHI6UP} or True {$ENDIF} then begin
    DefaultDraw := False;
//    if FCommonData.FUpdating then Exit; Black rects blinking avoiding
    if not (Stage in [cdPrePaint]) then
      DefaultDraw := False
    else
      if Stage in [cdPrePaint] then begin
        if not (csDesigning in ComponentState) then begin
          if not Flat and (HotButtonIndex = Button.Index) then
            State := State + [cdsHot];

          if DroppedButton = Button then
            State := State + [cdsHot];
        end;

        Flags := Flags + [tbNoEtchedEffect, tbNoEdges];
        iR := GetButtonRect(Button.Index);
        if (WidthOf(iR) <> Button.Width) or (HeightOf(iR) <> Button.Height) then begin
          Loaded; // Reinit buttons
          iR := GetButtonRect(Button.Index);
        end;
        if WidthOf(iR) <> 0 then begin
          bWidth := WidthOf(iR, True);
          bHeight := HeightOf(iR, True);
          BtnBmp := CreateBmp32(bWidth, bHeight);
          BtnBmp.Canvas.Font.Assign(Font);
          if not Button.Marked and not Button.Indeterminate and ((State = []) or (State = [cdsDisabled])) then
            Mode := 0
          else
            Mode := 1 + integer(([cdsSelected, cdsChecked] * State <> []) or Button.Marked or Button.Indeterminate);

{$IFNDEF DELPHI6UP}
          if FCommonData.SkinManager <> nil then
{$ENDIF}
{$IFDEF DELPHI6UP}
            if Menu <> nil then
              iSkinIndex := FCommonData.SkinManager.SkinCommonInfo.Sections[ssMenuItem]
            else
{$ENDIF}
              if Flat then
                iSkinIndex := FCommonData.SkinManager.SkinCommonInfo.Sections[ssToolButton]
              else
                iSkinIndex := FCommonData.SkinManager.SkinCommonInfo.Sections[ssSpeedButton]
{$IFNDEF DELPHI6UP}
          else begin
            iSkinIndex := -1;
            ci := EmptyCI;
          end
{$ENDIF}
          ;
          if iSkinIndex >= 0 then begin
            ci := MakeCacheInfo(FCommonData.FCacheBmp,
                                BorderWidth * 2 + integer(ebLeft in EdgeBorders) * (integer(EdgeInner <> esNone) + integer(EdgeOuter <> esNone)),
                                BorderWidth * 2 + integer(ebTop  in EdgeBorders) * (integer(EdgeInner <> esNone) + integer(EdgeOuter <> esNone)));

            R := MkRect(bWidth, Button.Height);
            OffsetRect(R, ClientRect.Left, ClientRect.Top);

            PaintItemBg(iSkinIndex, ci, Mode, R, Point(Button.Left, Button.Top), BtnBmp, FCommonData.CommonSkinData);
            R.Right := bWidth - AddedWidth;

            ci.X := ci.X + Button.Left;
            ci.Y := ci.Y + Button.Top;
            if FCommonData.CommonSkinData.gd[iSkinIndex].BorderIndex >= 0 then
              DrawSkinRect(BtnBmp, R, ci, FCommonData.CommonSkinData.ma[FCommonData.CommonSkinData.gd[iSkinIndex].BorderIndex], Mode, True);
{$IFNDEF DELPHI6UP}
          end
          else begin
            FillDC(BtnBmp.Canvas.Handle, MkRect(BtnBmp), clBtnFace);
            StdPaint(Button, State, BtnBmp);
{$ENDIF}
          end;
          DrawBtnCaption;
          DrawBtnGlyph;
          if Button.Style = tbsDropDown then
            DrawArrow;

          if not Button.Enabled or Button.Indeterminate then
            BmpDisabledKind(BtnBmp, FDisabledKind, Parent, CI, MkPoint);

          BitBlt(Canvas.Handle, Button.Left, Button.Top, bWidth, bHeight, BtnBmp.Canvas.Handle, 0, 0, SRCCOPY);

          FreeAndNil(BtnBmp);
        end;
      end;
  end
  else begin
    DefaultDraw := True;
    inherited;
  end;
end;


procedure UpdateInvalRects(SkinData: TsCommonData; State: integer);
var
  CurrentBounds: TRect;
  bWidth: integer;
begin
  if SkinData.FOwnerControl <> nil then begin
    CurrentBounds := MkRect(SkinData.FOwnerControl);
    // InvalidRectH
    if (CurrentBounds.Right = SkinData.FCacheBmp.Width) or
         (SkinData.CommonSkinData.gd[SkinData.SkinIndex].Props[State].GradientPercent <> 0) and
           (SkinData.BGType and BGT_GRADIENTHORZ <> 0) then
      // If Rect is not used
      SkinData.InvalidRectH.Right := 0
    else begin
      if SkinData.CommonSkinData.gd[SkinData.SkinIndex].BorderIndex < 0 then
        bWidth := 0
      else
        bWidth := SkinData.CommonSkinData.ma[SkinData.CommonSkinData.gd[SkinData.SkinIndex].BorderIndex].WR;

      if CurrentBounds.Right > SkinData.FCacheBmp.Width then begin
        SkinData.InvalidRectH.Left := SkinData.FCacheBmp.Width - bWidth;
        SkinData.InvalidRectH.Right := CurrentBounds.Right;
      end
      else begin
        SkinData.InvalidRectH.Left := CurrentBounds.Right - bWidth;
        SkinData.InvalidRectH.Right := CurrentBounds.Right;
      end;
      SkinData.InvalidRectH.Top := 0;
      SkinData.InvalidRectH.Bottom := CurrentBounds.Bottom;
    end;
    // InvalidRectV
    if (CurrentBounds.Bottom = SkinData.FCacheBmp.Height)
         or (SkinData.CommonSkinData.gd[SkinData.SkinIndex].Props[State].GradientPercent <> 0)
           and (SkinData.BGType and BGT_GRADIENTVERT <> 0) then
      // If Rect is not used
      SkinData.InvalidRectV.Bottom := 0
    else begin
      if SkinData.CommonSkinData.gd[SkinData.SkinIndex].BorderIndex < 0 then
        bWidth := 0
      else
        bWidth := SkinData.CommonSkinData.ma[SkinData.CommonSkinData.gd[SkinData.SkinIndex].BorderIndex].WB;

      if CurrentBounds.Bottom > SkinData.FCacheBmp.Height then begin
        SkinData.InvalidRectV.Top := SkinData.FCacheBmp.Height - bWidth;
        SkinData.InvalidRectV.Bottom := CurrentBounds.Bottom;
      end
      else begin
        SkinData.InvalidRectV.Top := CurrentBounds.Bottom - bWidth;
        SkinData.InvalidRectV.Bottom := CurrentBounds.Bottom;
      end;
      SkinData.InvalidRectV.Left := 0;
      SkinData.InvalidRectV.Right := CurrentBounds.Right;
    end;
  end;
end;


function TsToolBar.PrepareCache: boolean;
var
  ParentBG: TacBGInfo;
begin
  ParentBG.PleaseDraw := False;
  GetBGInfo(@ParentBG, Parent);
  if ParentBG.BgType = btNotReady then begin
    SkinData.Updating := True;
    Result := False;
  end
  else begin
    UpdateInvalRects(FCommonData, 0);
    InitCacheBmp(SkinData);
    PaintItem(FCommonData, BGInfoToCI(@ParentBG), True, 0, MkRect(Self), Point(Left, Top), FCommonData.FCacheBmp, True);
    FCommonData.BGChanged := False;
    Result := True;
  end;
end;


procedure TsToolBar.RepaintButton(Index: integer);
var
  Flags: TTBCustomDrawFlags;
  DC, SavedDC: HDC;
  RC, RW: TRect;
  Def: boolean;
begin
  if (Index >= 0) and Buttons[Index].Visible then begin
    Flags := [tbNoEtchedEffect, tbNoEdges];
    Def := False;
    DC := GetWindowDC(Handle);
    SavedDC := SaveDC(DC);
    try
      Windows.GetClientRect(Handle, RC);
      GetWindowRect(Handle, RW);
      MapWindowPoints(0, Handle, RW, 2);
      OffsetRect(RC, -RW.Left, -RW.Top);
      MoveWindowOrg(DC, -RW.Left, -RW.Top);

      Canvas.Handle := DC;
      OurAdvancedCustomDrawButton(Self, Buttons[Index], [], cdPrePaint, Flags, Def)
    finally
      Canvas.Handle := 0;
      RestoreDC(DC, SavedDC);
      ReleaseDC(Handle, DC);
    end;
  end;
end;


procedure TsToolBar.SetDisabledKind(const Value: TsDisabledKind);
begin
  if FDisabledKind <> Value then begin
    FDisabledKind := Value;
    Repaint;
  end;
end;


function TsToolBar.TextIsVisible(Button: TToolButton): boolean;
begin
  Result := ShowCaptions {$IFDEF D2010}and not (AllowTextButtons and List and (Button.Style <> tbsTextButton)){$ENDIF} {$IFDEF D2010}or (AllowTextButtons and (Button.Style = tbsTextButton)) {$ENDIF};
end;


procedure TsToolBar.UpdateEvents;
begin
{$IFDEF DELPHI6UP}
  if SkinData.Skinned then begin
{$ENDIF}
    inherited OnAdvancedCustomDraw := OurAdvancedCustomDraw;
    if not (csDesigning in ComponentState) then
      inherited OnAdvancedCustomDrawButton := OurAdvancedCustomDrawButton;
{$IFDEF DELPHI6UP}
  end
  else begin
    inherited OnAdvancedCustomDraw := nil;
    if not (csDesigning in ComponentState) then
      inherited OnAdvancedCustomDrawButton := nil;
  end;
{$ENDIF}
end;


procedure TsToolBar.WMNCPaint(var Message: TWMNCPaint);
var
  DC, SavedDC: HDC;
  RC, RW: TRect;
  w, h: integer;
begin
  if HandleAllocated and IsWindowVisible(Handle) then
    if FCommonData.Skinned then begin
      if not InUpdating(FCommonData) and not (csDestroying in ComponentState) then begin
        if not (FCommonData.BGChanged or FCommonData.FCacheBmp.Empty) or PrepareCache then begin
          DC := GetWindowDC(Handle);
          SavedDC := SaveDC(DC);
          try
            Windows.GetClientRect(Handle, RC);
            GetWindowRect(Handle, RW);
            MapWindowPoints(0, Handle, RW, 2);
            OffsetRect(RC, -RW.Left, -RW.Top);
            ExcludeClipRect(DC, RC.Left, RC.Top, RC.Right, RC.Bottom);
            { Draw borders in non-client area }
            w := WidthOf(Rc);
            h := HeightOf(Rc);
            // Top
            BitBlt(DC, 0, 0, Width, Rc.Top, SkinData.FCacheBmp.Canvas.Handle, 0, 0, SRCCOPY);
            // Left
            BitBlt(DC, 0, Rc.Top, Rc.Left, h, SkinData.FCacheBmp.Canvas.Handle, 0, Rc.Top, SRCCOPY);
            // Bottom
            BitBlt(DC, 0, Rc.Bottom, Width, Height - h - Rc.Top, SkinData.FCacheBmp.Canvas.Handle, 0, Rc.Bottom, SRCCOPY);
            // Right
            BitBlt(DC, Rc.Right, Rc.Top, Width - Rc.Left - w, h, SkinData.FCacheBmp.Canvas.Handle, Rc.Right, Rc.Top, SRCCOPY);

            IntersectClipRect(DC, RW.Left, RW.Top, RW.Right, RW.Bottom);
          finally
            RestoreDC(DC, SavedDC);
            ReleaseDC(Handle, DC);
          end;
        end;
      end;
    end
    else
      inherited;
end;


procedure TsToolBar.WndProc(var Message: TMessage);
var
  i, OldIndex, w, h: integer;
  PS: TPaintStruct;
  rc: TRect;
  DC: hdc;
begin
{$IFDEF LOGGED}
  AddToLog(Message);
{$ENDIF}
  if Message.Msg = SM_ALPHACMD then begin
    case Message.WParamHi of
      AC_CTRLHANDLED: begin
        Message.Result := 1;
        Exit;
      end;

      AC_SETNEWSKIN:
        if ACUInt(Message.LParam) = ACUInt(SkinData.SkinManager) then begin
          CommonWndProc(Message, FCommonData);
          AlphaBroadCast(Self, Message);
          UpdateEvents;
          Exit;
        end;

      AC_ENDPARENTUPDATE:
        if FCommonData.FUpdating then begin
          if not InUpdating(FCommonData, True) then begin
            FCommonData.BGChanged := True;
            if HandleAllocated then
              RedrawWindow(Handle, nil, 0, RDWA_NOCHILDRENNOW);
          end;
          SetParentUpdated(Self);
          Exit;
        end;

      AC_REFRESH:
        if RefreshNeeded(SkinData, Message) then begin
          DroppedButton := nil;

          CommonWndProc(Message, FCommonData);
          AlphaBroadcast(Self, Message);
          UpdateEvents;

          if IsWindowVisible(Handle) then begin
            SendMessage(Handle, WM_NCPAINT, 0, 0);
            Repaint;
          end;
          Exit;
        end;

      AC_REMOVESKIN:
        if ACUInt(Message.LParam) = ACUInt(SkinData.SkinManager) then begin
          DroppedButton := nil;
          CommonWndProc(Message, FCommonData);
          AlphaBroadcast(Self, Message);
          UpdateEvents;
          if IsWindowVisible(Handle) then begin
            Repaint;
            SendMessage(Handle, WM_NCPAINT, 0, 0);
          end;
          Exit;
        end;

      AC_GETDEFSECTION: begin
        Message.Result := 3 + 1;
        Exit;
      end;

      AC_GETDEFINDEX: begin
        if FCommonData.SkinManager <> nil then
          Message.Result := FCommonData.SkinManager.SkinCommonInfo.Sections[ssToolBar] + 1;

        Exit;
      end;
    end;
  end;

  if not ControlIsReady(Self) or not FCommonData.Skinned then
    inherited
  else begin
    if Message.Msg = SM_ALPHACMD then
      case Message.WParamHi of
        AC_GETBG:
          with Message, PacBGInfo(LParam)^ do begin
            Offset := MkPoint;
            InitBGInfo(FCommonData, PacBGInfo(LParam), 0);
            inc(Offset.X, OffsetX);
            inc(Offset.Y, OffsetY);
            if (BgType = btFill) and (Bmp <> nil) then begin
              FillRect.Left := Offset.X - FillRect.Left;
              FillRect.Top  := Offset.Y - FillRect.Top;
            end;
            Exit;
          end

        else
          if CommonMessage(Message, SkinData) then
            Exit;
      end
    else
      case Message.Msg of
        WM_PRINT:
          with TWMPaint(Message) do begin
            FCommonData.Updating := False;
            RC := ACClientRect(Handle);
            w := WidthOf(Rc);
            h := HeightOf(Rc);
            PrepareCache;
            BitBlt(DC, 0, 0, Width, Rc.Top, SkinData.FCacheBmp.Canvas.Handle, 0, 0, SRCCOPY);
            BitBlt(DC, 0, Rc.Top, Rc.Left, h, SkinData.FCacheBmp.Canvas.Handle, 0, Rc.Top, SRCCOPY);
            BitBlt(DC, 0, Rc.Bottom, Width, Height - h - Rc.Top, SkinData.FCacheBmp.Canvas.Handle, 0, Rc.Bottom, SRCCOPY);
            BitBlt(DC, Rc.Right, Rc.Top, Width - Rc.Left - w, h, SkinData.FCacheBmp.Canvas.Handle, Rc.Right, Rc.Top, SRCCOPY);
            MoveWindowOrg(DC, Rc.Left, Rc.Top);
            IntersectClipRect(DC, 0, 0, w, h);
            if SkinIsChanging(SkinData) then begin // Preventing of blinking when skin is changed
              SendMessage(Parent.Handle, WM_SETREDRAW, 0, 0);
              SendMessage(Handle, WM_PAINT, Message.WParam, 0);
              SendMessage(Parent.Handle, WM_SETREDRAW, 1, 0);
            end
            else
              SendMessage(Handle, WM_PAINT, Message.WParam, 0);

            for i := 0 to ButtonCount - 1 do
              with Buttons[i] do
                if (Style in [tbsSeparator, tbsDivider]) and Visible and not Wrap then begin
                  rc := BoundsRect;
                  PaintSeparator(DC, rc, Self, (Style = tbsDivider));
                end;

            Exit;
          end;

        WM_ERASEBKGND:
          if ([csDesigning, csDestroying] * ComponentState = []) and not InUpdating(FCommonData) then begin
            FCommonData.Updating := FCommonData.Updating;
            Message.Result := 1;
            Exit;
          end;

        CM_COLORCHANGED:
          Exit;

        CM_TEXTCHANGED: begin
          Repaint;
          Exit;
        end;
{
        TBN_DROPDOWN: begin
          NMTB := NMToolBar;
          with NMTB^ do
            if Perform(TB_GETBUTTON, iItem, tbButton) <> 0 then
              DroppedButton := TToolButton(tbButton.dwData);
        end;
}
        CN_NOTIFY:
          with TWMNotify(Message) do
            case TWMNotify(Message).NMHdr^.code of
              TBN_DROPDOWN:
                with PNMToolBar(NMHdr)^ do
                  if Perform(TB_GETBUTTON, iItem, LPARAM(@tbButton)) <> 0 then begin
                    DroppedButton := TToolButton(tbButton.dwData);
                    DroppedButton.Repaint
                  end;

              TBN_DELETINGBUTTON:
                if HotButtonIndex >= ButtonCount then
                  HotButtonIndex := -1;
            end;

        WM_PAINT:
          if InAnimationProcess and (0 = SkinData.PrintDC) then
            Exit
          else
            if InUpdating(SkinData) then begin
              BeginPaint(Handle, PS);
              EndPaint(Handle, PS);
              Exit;
            end;

{$IFDEF DELPHI7UP}
        WM_LBUTTONDBLCLK:
          if Customizable then // Remove a custom item drawing for avoiding a bug in VCL
            if not (csDesigning in ComponentState) then
              inherited OnAdvancedCustomDrawButton := nil;
{$ENDIF}
      end;

    if CommonWndProc(Message, FCommonData) then
      Exit;
      
    inherited;
    case Message.Msg of
{$IFDEF DELPHI7UP}
      WM_LBUTTONDBLCLK:
        if Customizable then // Restore a custom drawing
          UpdateEvents;
{$ENDIF}

      WM_PAINT:
        if not InAnimationProcess then begin
          if TWMPAint(Message).DC = 0 then
            DC := GetWindowDC(Handle)
          else
            DC := TWMPAint(Message).DC;

          w := SkinData.SkinManager.SkinCommonInfo.Sections[ssDivider];
          if SkinData.SkinManager.IsValidSkinIndex(w) then
            for i := 0 to ButtonCount - 1 do
              with Buttons[i] do
                if (Style in [tbsSeparator, tbsDivider]) and Visible then
                  if not Wrap then begin
                    rc := BoundsRect;
                    PaintSeparator(DC, rc, Self, (Style = tbsDivider));
                  end;

          if TWMPAint(Message).DC = 0 then
            ReleaseDC(Handle, DC);
        end;

      CN_DROPDOWNCLOSED:
        if (DroppedButton <> nil) and (DroppedButton.Style = tbsDropDown) then begin
          HotButtonIndex := -1;
          i := DroppedButton.Index;
          DroppedButton := nil;
          RepaintButton(i);
        end;

      CM_MOUSELEAVE:
        if not Flat and not (csDesigning in ComponentState) then begin
          OldIndex := HotButtonIndex;
          HotButtonIndex := -1;
          if (OldIndex >= 0) and not Buttons[OldIndex].Down then
            RepaintButton(OldIndex);

          HotButtonIndex := -1;
        end;

      WM_MOUSEMOVE:
        if not Flat and not (csDesigning in ComponentState) then
          with TWMMouse(Message) do begin
            i := IndexByMouse(Point(XPos, YPos));
            if i <> HotButtonIndex then begin
              if (i >= 0) and not Buttons[i].Enabled then
                Exit;

              OldIndex := HotButtonIndex;
              HotButtonIndex := i;
              if (OldIndex >= 0) and not Buttons[OldIndex].Down then
                RepaintButton(OldIndex);

              if (HotButtonIndex >= 0) and not Buttons[HotButtonIndex].Down then
                RepaintButton(HotButtonIndex);
            end;
          end;

      WM_SIZE, CM_VISIBLECHANGED, CM_ENABLEDCHANGED: begin
//        FCommonData.BGChanged := True;
//        InvalidateRect(Handle, nil, False);
        FCommonData.Invalidate(True);
      end;
    end;
  end;
end;

end.

